import numpy as np
import matplotlib.pyplot as plt


def draw_ellipsoid(rx, ry, rz, plotFig=True, res=100):
    """
    Calculate and draw ellipsoid

    Parameters
    ----------
    rx : float
        Radius of the ellipsoid in x.
    ry : float
        Radius of the ellipsoid in y.
    rz : float
        Radius of the ellipsoid in z.
    plotFig : boolean, optional
        Plot figure [True / False]. The default is True.
    res : int, optional
        Resolution, i.e. number of data points. The default is 100.

    Returns
    -------
    [x, y, z]   List of x, y, z coordinates of the ellipsoid
    Figure      Plot with ellipsoid if plotFig==True.

    """
    
    # Set of all spherical angles:
    u = np.linspace(0, 2 * np.pi, res)
    v = np.linspace(0, np.pi, res)
    
    # Cartesian coordinates that correspond to the spherical angles:
    # (this is the equation of an ellipsoid):
    x = rx * np.outer(np.cos(u), np.sin(v))
    y = ry * np.outer(np.sin(u), np.sin(v))
    z = rz * np.outer(np.ones_like(u), np.cos(v))
    
    # Plot:
    if plotFig:
        fig = plt.figure()
        ax = plt.axes(projection = "3d")
        ax.plot_surface(x, y, z,  rstride=4, cstride=4, color='b', alpha=0.2)
        Min = np.min([np.min(x), np.min(y), np.min(z)])
        Max = np.max([np.max(x), np.max(y), np.max(z)])
        ax.set_xlim([Min, Max])
        ax.set_ylim([Min, Max])
        ax.set_zlim([Min, Max])
    
    return([x, y, z])



import numpy as np
from ..tools.phasor import phasor


def data2phasor(data):
    """
    Calculate phasors for all histograms in data object

    Parameters
    ----------
    data : object
        Object with histograms for each channel.
        E.g. data.hist1 = np.array(N x 2)

    Returns
    -------
    data : object
        Same object with phasors included.

    """
    
    histList = [i for i in list(data.__dict__) if i.startswith('hist')]
    N = len(histList)
    
    phasors = np.zeros(N, dtype=complex)
    
    for i in range(N):
        hist = getattr(data, histList[i])
        hist = hist[:,1]
        phasors[i] = phasor(hist)
    
    data.phasors = phasors
    
    return data


import numpy as np
from ..tools.savevar import openvar


def atimes2timetrace(ArrivalTimes):
    """
    Convert a list of ArrivalTimes to an intensity time trace I(t) for each of
    the detector elements of the SPAD array detector

    Parameters
    ----------
    ArrivalTimes : object
        Variable with the ArrivalTimes for each detector
            element (in units of dwellTime)
            --> output of binfile2arrivaltimes.py.

    Returns
    -------
    data : np.array()
        N x 26 array with for each row the number of photon counts
            for each channel and the sum of all channels per microbin.

    """

    # Number of time bins
    N = ArrivalTimes.sum[-1] + 1

    # Allocate data array
    data = np.zeros((N, 26), dtype='H')

    # Detector elements 0-24
    for det in range(25):
        # Create empty time trace
        timeTrace = np.zeros((N, 1), dtype='H')
        singleCh = getattr(ArrivalTimes, 'det' + str(det))
        print('Det' + str(det) + ' -> ' + str(len(singleCh)) + ' photons')
        if len(singleCh) != len(set(singleCh)):
            # there exists a time bin with more than one photon in a single bin
            for i in range(len(singleCh)):
                data[singleCh[i], det] += 1
        elif len(singleCh) == 0:
            # no photons in this channel
            # do nothing
            pass
        else:
            timeTrace[singleCh, 0] = 1
            data[:, det] = timeTrace

    # Sum of all elements
    timeTrace = np.zeros((N, 1), dtype='H')
    singleCh = getattr(ArrivalTimes, 'sum')
    print('Sum over all channels' + ' -> ' + str(len(singleCh)) + ' photons')
    if len(singleCh) != len(set(singleCh)):
        # there exists a time bin with more than one photon in a single bin
        for i in range(len(singleCh)):
            data[singleCh[i], 25] += 1
    else:
        timeTrace[singleCh, 0] = 1
        data[:, 25] = timeTrace

    return data


def atimes2data(fname):
    """
    Import .pickle file with arrival time data for each channel and convert to
    normal data array

    Parameters
    ----------
    fname : string
        file name.

    Returns
    -------
    data : np.array()
        data array.

    """
    
    aTimes = openvar(fname)
    data = atimes2timetrace(aTimes)
    return data


def atimes2timetrace_bh(ArrivalTimes, bin_length):
    """
    Convert a list of arrivalTimes to an intensity time trace I(t)

    Parameters
    ----------
    arrivalTimes : object
        Variable with the arrivalTimes for each detector
            element [of ns].
    binLength : float
        Duration of each bin [in ns].

    Returns
    -------
    data : np.array()
        Vector with intensity trace vs time in bins of binLength.

    """
    
    # calculate for each photon in which bin it should be
    photonBins = np.int64(ArrivalTimes / bin_length)
    
    # number of photon bins
    Nbins = np.max(photonBins) + 1
    
    # create output vector
    data = np.zeros(Nbins, 'int16')
    for i in range(len(photonBins)):
        data[photonBins[i]] += 1
        
    return data



from os.path import split

def path2fname(path, properWay=False):
    """
    Split path into file name and folder

    Parameters
    ----------
    path : path
        Path to a file [string], e.g. 'C:\\Users\\SPAD-fcs\\file.ext'.
    properWay : TYPE, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    fname : str
        File name, e.g. 'file.ext'.
    folderName : path
        Folder name, e.g. 'C:\\Users\\SPAD-fcs\\'.

    """
    
    if properWay:
        output = split(path)
        folderName = output[0]
        fname = output[1]
    else:
        path = path.replace("/", "\\")
        fname = path.split('\\')[-1]
        folderName = '\\'.join(path.split('\\')[0:-1]) + '\\'
    
    return fname, folderName



import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import least_squares

def fit_psf(image, pxsize=1, fitRange=15, Nbeads=5):
    """
    Show image of beads. The user selects 5 beads. Gaussians are fitted to get
    the width/height of the PSF.

    Parameters
    ----------
    image : np.array()
        2D/3D array with image data.
    pxsize : float, optional
        pixel size (currently not used). The default is 1.
    fitRange : int, optional
        number of pixels up, down, left and right used as fit region. The default is 15.
    Nbeads : int, optional
        Number of beads. The default is 5.

    Returns
    -------
    listOfResults : list
        List of the fit results, width is 1/e^2 radius.
        [x0, y0, z0, wx, wy, wz, ax, ay, A, offset]

    """
    
    
    # check if data is 2D or 3D
    Ndim = len(np.shape(image))
    
    # sum all z positions for 3D data
    if Ndim == 2:
        imageSum = image
    else:
        imageSum = np.sum(image, 2)
    
    # plot image
    plt.figure()
    plt.imshow(imageSum)
    dataPoints = plt.ginput(Nbeads, show_clicks=True)
    
    # store results
    if Ndim == 2:
        listOfResults = np.zeros([Nbeads, 7])
    else:
        listOfResults = np.zeros([Nbeads, 10])

    for i in range(Nbeads):
        print(i)
        # go through all chosen points and fit 2D/3D Gaussian around this
        # center coordinates of the Gaussian
        x0 = int(dataPoints[i][0])
        y0 = int(dataPoints[i][1])
        # check region around center to not be out of bounds
        xmin, xmax = checkrange(x0, fitRange, 0, np.size(image, 1))
        ymin, ymax = checkrange(y0, fitRange, 0, np.size(image, 0))
        # get local image
        imageLoc = imageSum[ymin:ymax, xmin:xmax]
        plt.figure()
        plt.imshow(imageLoc)
        # fit Gaussian
        if Ndim == 2:
            imageLoc = image[ymin:ymax, xmin:xmax]
            res_lsq = gauss_fitres(imageLoc, np.array([fitRange, fitRange+1, 3, 3, 1, 0, 0.01]))
            # x0, y0, wx, wy, A, offset
            print(res_lsq.x)
            plt.figure()
            plt.imshow(gaussian(np.size(image, 1), np.size(image, 0), res_lsq.x[0] + xmin, res_lsq.x[1] + ymin, res_lsq.x[2], res_lsq.x[3], res_lsq.x[4], res_lsq.x[5], res_lsq.x[6]))
        else:
            Nz = np.shape(image)
            Nz = Nz[2]
            imageLoc = image[ymin:ymax, xmin:xmax, :]
            res_lsq = gauss_fitres3d(imageLoc, [fitRange, fitRange+1, int(Nz/2), 3, 3, 3, 0.1, 0.1, 1, 0]) # [x0, y0, z0, wx, wy, wz, ax, ay, A, offset]
        # print results
        print(res_lsq.success)
        print(res_lsq.x)
        if res_lsq.success:
            listOfResults[i, :] = np.abs(res_lsq.x)
    print('---------------------------------------')
    print('Average')
    print('---------------------------------------')
    print(np.average(listOfResults, 0))
    print('---------------------------------------')
    return listOfResults
        

def checkrange(x, fitRange, xmin, xmax):
    """
    Check that the region around the center of a Gaussian does not go out of
    bounds. If the user selects a bead close to the border of the image, the
    region around it, which is needed to fit the data, may go out of bounds.
    This function cuts the range to 0 and N with N+1 the width/height of the
    image.

    Parameters
    ----------
    x : float
        bead coordinate (either x or y).
    fitRange : list
        number of pixels up, down, left and right used as fit region.
    xmin : int
        minimum coordinate value allowed (0) before going of out bounds.
    xmax : int
        maximuum coordinate value allowed (N).

    Returns
    -------
    xmin, xmax : int
        Input variables, but cut off if needed
                    E.g. checkrange(10, 15, 0, 100) returns [0, 25]:
                        10 - 15 --> 0
                        10 + 15 --> 25.
    
    """
    
    xmin = np.max([x - fitRange, xmin])
    xmax = np.min([x + fitRange, xmax-1])
    return xmin, xmax
        
        
def gauss_fitres(image, param, fitinfo=np.array([1, 1, 1, 1, 1, 1, 1])):
    """
    2D Gaussian least squares fit

    Parameters
    ----------
    image : np.array
        2D matrix with image data to fit.
    param : np.array
        np.array with initialization values for the fit parameters
        see gaussfitfun2d for more information
        order: [x0, y0, wx, wy, A, offset, rot].
    fitinfo : np.array, optional
        Boolean np.array with 1 for fitted parameters and 0 for fixed
        parameters. E.g. to fit only x0 and y0
        fitinfo = [1, 1, 0, 0, 0, 0, 0].
        The default is np.array([1, 1, 1, 1, 1, 1, 1]).

    Returns
    -------
    fitresult : object
        Object with the results of the fit.

    """
    
    fitparamStart = param[fitinfo==1]
    fixedparam = param[fitinfo==0]
    
    fitresult = least_squares(gaussfitfun2d, fitparamStart, args=(fixedparam, fitinfo, image))
    return fitresult


def gauss_fitres3d(image, fitparamStart):
    """
    3D Gaussian least squares fit

    Parameters
    ----------
    image : np.array()
        3D matrix with image data to fit.
    fitparamStart : np.array()
        Vector with initialization values for the fit parameters
        see gaussfitfun3d for more information.

    Returns
    -------
    fitresult : object
        Object with the results of the fit.

    """
    
    fitresult = least_squares(gaussfitfun3d, fitparamStart, args=(fitparamStart[0], image))
    return fitresult
        

def gaussfitfun2d(fitparamStart, fixedparam, fitinfo, image):
    """
    Calculate residuals of 2D Gaussian fit of 2D data

    Parameters
    ----------
    fitparamStart : np.array()
        Vector with initialization values for the fitted parameters.
    fixedparam : np.array()
        Vector with the values for the fixed parameters
        order for both vectors: [x0, y0, wx, wy, A, offset, rot]
        x0, y0      center coordinates of the Gaussian function
        wx, wy      1/e^2 width in x and y
        A           amplitude of the Gaussian
        offset      offset of the Gaussian.
    fitinfo : np.array()
        Boolean vector of length 7 with 1's for the fitted parameters.
    image : np.array()
        2D matrix with image data to fit.

    Returns
    -------
    res : np.array()
        1D array with all the residuals.

    """
    
    Ny = np.size(image, 0)
    Nx = np.size(image, 1)
    
    fitparam = np.float64(np.zeros(7))
    fitparam[fitinfo==1] = fitparamStart
    fitparam[fitinfo==0] = fixedparam
    
    x0 = fitparam[0]
    y0 = fitparam[1]
    wx = fitparam[2]
    wy = fitparam[3]
    A = fitparam[4]
    offset = fitparam[5]
    rot = fitparam[6]
    
    res = np.reshape(gaussian(Nx, Ny, x0, y0, wx, wy, A, offset, rot) - image, [Nx*Ny, 1])
    res = np.ravel(res)
    
    return res


def gaussfitfun3d(fitparam, dummy, image):
    """
    Calculate residuals of 3D Gaussian fit of 3D data

    Parameters
    ----------
    fitparam : np.array()
        Vector with initialization values for the fit parameters
        [x0, y0, z0, wx, wy, wz, ax, ay, A, offset]
        x0, y0, z0  center coordinates of the Gaussian function
        wx, wy, zw  1/e^2 width in x, y, and z
        ax, ay      skew in x and y
        A           amplitude of the Gaussian
        offset      offset of the Gaussian.
    dummy : TYPE
        Not used.
    image : np.array()
        3D matrix with image data to fit.

    Returns
    -------
    res : np.array()
        1D array with all the residuals.

    """
    
    Ny = np.size(image, 0)
    Nx = np.size(image, 1)
    Nz = np.size(image, 2)
    
    x0 = fitparam[0]
    y0 = fitparam[1]
    z0 = fitparam[2]
    wx = fitparam[3]
    wy = fitparam[4]
    wz = fitparam[5]
    ax = fitparam[6]
    ay = fitparam[7]
    A = fitparam[8]
    offset = fitparam[9]
    
    res = np.reshape(gaussian3d(Nx, Ny, Nz, x0, y0, z0, wx, wy, wz, ax, ay, A, offset) - image, [Nx*Ny*Nz, 1])
    res = np.ravel(res)
    
    return res


def gaussian(Nx, Ny, x0, y0, wx, wy, A, offset, rot):
    """
    Calculate 2D Gaussian function

    Parameters
    ----------
    Nx : int
        Number of data points in horizontal, vertical direction.
    Ny : int
        Number of data points in horizontal, vertical direction.
    x0 : int
        Horizontal, vertical center of the Gaussian function
        (integers representing column and row).
    y0 : int
        Horizontal, vertical center of the Gaussian function
        (integers representing column and row).
    wx : float
        1/e^2 width of the Gaussian in horizontal, vertical direction.
    wy : float
        1/e^2 width of the Gaussian in horizontal, vertical direction.
    A : float
        Amplitude of the Gaussian.
    offset : float
        Offset.
    rot : float
        rotation.

    Returns
    -------
    out : np.array()
        2D matrix with a 2D Gaussian function.

    """
    
    meshgrid = np.meshgrid(np.linspace(0, Nx-1, Nx), np.linspace(0, Ny-1, Ny))
    x = meshgrid[0]
    y = meshgrid[1]
    out = A * np.exp(-2 * ((x-x0-rot*(y-y0))**2 / wx**2 + (y-y0)**2 / wy**2)) + offset
    return out


def gaussian3d(Nx, Ny, Nz, x0, y0, z0, wx, wy, wz, ax, ay, A, offset):
    """
    Calculate 3D Gaussian function

    Parameters
    ----------
    Nx, Ny, Nz : int
        Number of data points in horizontal, vertical, and z direction.
    x0, y0, z0 : float
        Horizontal, vertical, and z center of the Gaussian function
        (integers representing column, row, and plane).
    wx, wy, wz : float
        1/e^2 width of the Gaussian in horizontal, vertical, and z-
                    direction.
    ax, ay : float
        Skew in the x and y direction.
    A : float
        Amplitude of the Gaussian.
    offset : float
        Offset.

    Returns
    -------
    out : np.array()
        3D matrix with a 3D Gaussian function.

    """
    
    meshgrid = np.meshgrid(np.linspace(0, Nx-1, Nx), np.linspace(0, Ny-1, Ny), np.linspace(0, Nz-1, Nz))
    x = meshgrid[0]
    y = meshgrid[1]
    z = meshgrid[2]
    out = A * np.exp(-2 * ((x+ax*z-x0)**2 / wx**2 + (y+ay*z-y0)**2 / wy**2) + (z-z0)**2 / wz**2) + offset
    return out
    



import numpy as np
import os
from .u64_to_counts import u64_to_counts
from .read_binfile import read_binfile
from ..tools.savevar import savevar

class ArrivalTimes:
    pass

def binfile2data(fname, store_pickle=False):
    """
    Read a binary fcs data file, containing a list of U64 numbers that
    represent the photon counts per microtime bin of about 1 µs.

    Parameters
    ----------
    fname : string
        Binary file name.
    store_pickle : boolean, optional
        store data in file. The default is False.

    Returns
    -------
    countArray : 2D numpy array
        N x 26 array with for each row the number of photon counts
        for each channel and the sum of all channels per microbin.

    """

    readLength = 8 * 65536

    # Get file size [in bytes]
    fsize = os.path.getsize(fname)

    # Empty array to store photon counts
    countArray = np.zeros((int(fsize/2/8), 26), dtype='H')

    startPos = 0  # position in the .bin file to start reading
    row = 0  # row of countArray to write to

    # Go through the file, read a chunk of data, write to the array, and repeat
    while startPos < fsize:
        print(startPos/fsize*100)
        # Read part of the binary file
        data = read_binfile(fname, startPos, readLength)
        startPos += readLength

        # Calculate photon counts
        for i in range(len(data)):
            counts = u64_to_counts(data[i])
            # add counts to array
            countArray[int(np.floor(row/2))][:] = np.add(
                    countArray[int(np.floor(row/2))][:],
                    counts)
            row += 1

    print(100)
    
    if store_pickle:
        print("Storing .pickle file")
        savevar(countArray, fname[0:-4] + "_data")
        print(".pickle file stored")
    
    return countArray


def binfile2dataAle(fname, store_pickle=False):
    """
    Read a binary data file (different FPGA), containing a list of U32 numbers
    that represent photon arrival times
    (not in use any more, will be deleted in future)

    Parameters
    ----------
    fname : string
        Binary file name.
    store_pickle : TYPE, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    data : object.
        Class with 32 fields, each one containing a list of the arrival
        times of the photons in that channel. With the 5x5 SPAD array
        detector, only the 25 first fields are used, the others are
        empty. The arrival times are in units of 12.5 ns, i.e.
        multiply each list by 12.5 to get the arrival times in ns.

    """
    
    # Open .bin file
    print("Opening .bin file.")
    with open(fname, mode='rb') as file:
        rawdata = list(file.read())
    print("Binfile opened.")
    
    # Number of bytes per line
    NB = 4
    
    # Number of addresses
    NA = 8
    
    # number of data points
    N = len(rawdata) / NB  # 4 bytes per 32 bit number

    # convert list into list of lists with each sublist containing 4 numbers
    # rawdata2 = [rawdata[i:i+NB] for i in range(0, len(rawdata), NB)]

    # reverse order of the bytes in each list item
    # rawdata = [list(reversed(rawdata[i])) for i in range(0, len(rawdata))]

    # Create vector data with the macro arrival times
    data = ArrivalTimes()
    for i in range(NA):
        for j in range(NB):
            setattr(data, "pixel" + str(i*NB + j), [])
    
    # go through data set
    oldTimeArr = np.zeros(NA)  # keep track of time
    k = np.zeros(NA)  # keep track of number of macrotimes of each address
    for i in range(round(N)):
        if np.mod(i, 10000) == 0:
            print("{0:.3f} %".format(100 * i/N))
        start = 4 * i
        newRawDataPoint = rawdata[start:start+NB]
        newRawDataPoint = list(reversed(newRawDataPoint))
        address = getaddress(newRawDataPoint)
        time = gettime(newRawDataPoint)
        ch = getchannel(newRawDataPoint)
        # print("time: " + str(time) + " - oldTime: " + str(oldTime))
        # print(ch)
        if time < oldTimeArr[address]:
            k[address] += 1
        oldTimeArr[address] = time
        time = int(k[address] * 4096 + time)
        if ch > 0 and address >= 0:
            pixel = address * 4 + ch - 1
            getattr(data, "pixel" + str(pixel)).append(time)
    
    if store_pickle:
        print("Storing .pickle file")
        savevar(data, fname[0:-4] + "_data")
        print(".pickle file stored")
    
    return(data)
        

def getaddress(data):
    address = data[0]
    address = list('{0:08b}'.format(int(address)))
    address = address[0:3]
    address = int("".join(str(x) for x in address), 2)
    return address


def gettime(data):
    byte0 = data[3]
    byte0 = list('{0:08b}'.format(int(byte0)))
    byte1 = data[2]
    byte1 = list('{0:08b}'.format(int(byte1)))
    byte1 = byte1[4:8]
    time = byte1 + byte0
    time = int("".join(str(x) for x in time), 2)
    return time


def getchannel(data):
    byte2 = data[1]
    byte2 = list('{0:08b}'.format(int(byte2)))
    byte3 = data[0]
    byte3 = list('{0:08b}'.format(int(byte3)))
    ch1 = byte2[7]
    ch2 = byte2[6]
    ch3 = byte3[7]
    ch4 = byte3[6]
    channels = [ch1, ch2, ch3, ch4]
    channels = list(map(int, channels))
    if np.sum(channels) > 1 or np.sum(channels) == 0:
        return -1
    return channels.index(1) + 1


import matplotlib.pyplot as plt
import numpy as np
import seaborn

def color_from_map(value, startv=0, stopv=1, cmap='viridis'):
    if cmap != 'seaborn-colorblind':
        cmap = plt.get_cmap(cmap)
        
        try:
            N = np.size(cmap.colors, 0)
        except:
            N = 255
        idx = int(np.floor((value - startv) / (stopv - startv) * N))
        idx = np.clip(idx, 0, N*1)
        
        try:
            c = cmap.colors[idx]
        except:
            c = cmap(idx/N)[0:3]
        
        return c
    else:
        cmapArray = seaborn.color_palette('colorblind')
        N = len(cmapArray)
        idx = int(np.floor((value - startv) / (stopv - startv) * N))
        idx = np.clip(idx, 0, N*1)
        return cmapArray[idx]



import numpy as np


def find_nearest(array, value):
    """
    Find nearest value and index of value in array.
    function call: [value2, idx] = find_nearest(array, value)

    Parameters
    ----------
    array : np.array
        1D array of values to look into.
    value : float
        Number to find in the array.

    Returns
    -------
    value2 : float
        Value in array that is closest to value.
    idx : int
        Index of this value in the array.

    """
    
    array = np.asarray(array)
    
    idx = (np.abs(array - value)).argmin()
    value2 = array[idx]
    
    return value2, idx



"""
CorrelationFunction fields:
    mode        str             name of the correlation, e.g. 'Pair-correlation fcs'
    elements    list of str     output names of the correlations, e.g. ['central', 'sum3', 'sum5']
    listOfG     list of str     names of the correlations to calculate, e.g. ['central', 'sum3', 'sum5']
    shortlabel  str             short name of the correlation, e.g. 'PairCorr'
                                (currently not used)
    average     list of str     list with correlations to average
                                e.g. for pair-correlation fcs: ['7x8+12x13+17x18', '12x7+11x6+13x8', '7x6+12x11+17x16', '11x16+12x147+13x18']
    
"""

class CorrelationFunction():
    def __init__(self):
        self.mode = None
        self.elements = None
        self.list_of_g = None
        self.shortlabel = None
        self.average = None
    
    def set_params(self, params):
        self.mode          = params.get("mode")
        self.elements      = params.get("elements")
        self.list_of_g     = params.get("listOfG")
        self.shortlabel    = params.get("shortlabel")
        self.average       = params.get("average")


import h5py
from .meas_to_count import file_to_fcs_count

def fcsbin2h5(binfile, h5file):
    data = file_to_fcs_count(binfile)

    h5f = h5py.File(h5file, 'w')
    h5f.create_dataset('dataset_1', data=data, compression="gzip")
    
    h5f.close()


import matplotlib.pyplot as plt
import numpy as np
from .distance2detelements import distance2detelements
from ..tools.cast_data import cast_data
from brighteyes_ism.simulation.detector import hex_grid

def plot_fingerprint_airyscan(counts, plot=False):
    """
    Plot the airyscan fingerprint from a data set with the photon counts for
    each of the 32 detector elements

    Parameters
    ----------
    counts : np.array
        Array with the 32 photon counts of the airyscan.
        Channel numbers according to Zeiss (ch 0 = center)
    plot : boolean, optional
        Plot the finger print

    Returns
    -------
    hexb : np.array()
        Array for hexbin plotting

    """
    
    N = 11  # Number of elements
    s2 = hex_grid(N, np.linspace(0, N - 1, N))
    
    hexb = np.zeros((66))
    
    hexb[0:19] = 0
    hexb[20] = counts[27]
    hexb[21] = counts[28]
    hexb[22] = 0
    hexb[23] = 0
    hexb[24] = counts[26]
    hexb[25] = counts[14]
    hexb[26] = counts[15]
    hexb[27] = counts[16]
    hexb[28] = counts[29]
    hexb[29] = 0
    hexb[30] = counts[25]
    hexb[31] = counts[13]
    hexb[32] = counts[4]
    hexb[33] = counts[5]
    hexb[34] = counts[17]
    hexb[35] = counts[30]
    hexb[36] = counts[12]
    hexb[37] = counts[3]
    hexb[38] = counts[0]
    hexb[39] = counts[6]
    hexb[40] = counts[18]
    hexb[41] = counts[31]
    hexb[42] = counts[24]
    hexb[43] = counts[11]
    hexb[44] = counts[2]
    hexb[45] = counts[1]
    hexb[46] = counts[7]
    hexb[47] = counts[19]
    hexb[48] = counts[23]
    hexb[49] = counts[10]
    hexb[50] = counts[9]
    hexb[51] = counts[8]
    hexb[52] = counts[20]
    hexb[53] = 0
    hexb[54] = 0
    hexb[55] = 0
    hexb[56] = counts[22]
    hexb[57] = counts[21]
    hexb[58:65] = 0
    
    if plot:
        fig, ax = plt.subplots()
        ax.set_facecolor("black")
        plt.hexbin(s2[1], s2[0], C=hexb, gridsize=[6,5], cmap="inferno", linewidths=0.5, edgecolors="k")
        plt.colorbar(label="Cell Value")
        plt.xlim([-0.5,5.5])
        plt.ylim([2,8.5])
        plt.xticks([])
        plt.yticks([])
        ax.set_box_aspect(1)
    return s2, hexb

def plot_airy(data, show_perc=True, dtype='int64', normalize=False, savefig=0, vminmax = 'auto'):
    """
    Make finger print plot of SPAD-fcs data with 25 channels.

    Parameters
    ----------
    data : np.array()
        Nx26 or Nx25 array with the fcs data
        or data object with data.det0 etc. arrival times.
    show_perc : boolean, optional
        Show percentages. The default is True.
    dtype : string, optional
        Data type. The default is 'int64'.
    normalize : boolean, optional
        Convert total counts to average counts per bin if True. The default is False.
    savefig : int, optional
        Path to store figure. The default is 0.
    vminmax : vector or string, optional
        Vector with minimum and maximum color bar value. The default is 'auto'.

    Returns
    -------
    airy : np.array()
        26 element vector with the sum of the rows and plot.

    """
    
    if type(data) == np.ndarray:
        # data is numpy array with intensity traces
        if len(np.shape(data)) > 1:
            # if 2D array, convert to dtype and sum over all rows
            data = cast_data(data, dtype)
            airy = np.sum(data, axis=0)
        else:
            airy = data
        airy2 = airy[0:25]
    else:
        if hasattr(data, 'det24'):
            # data is fcs2arrivaltimes.ATimesData object with 25 elements
            airy2 = np.zeros(25)
            for det in range(25):
                airy2[det] = len(getattr(data, 'det' + str(det)))
        else:
            # data is fcs2arrivaltimes.ATimesData object with 21 elements
            airy2 = np.zeros(25)
            # *  0  1  2  *
            # 3  4  5  6  7
            # 8  9  10 11 12
            # 13 14 15 16 17
            # *  18 19 20 *
            dets = [1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 21, 22, 23]
            for det in range(21):
                if hasattr(data, 'det' + str(det)):
                    airy2[dets[det]] = len(getattr(data, 'det' + str(det)))
                else:
                    airy2[dets[det]] = 0
        airy = airy2
    
    if normalize:
        airy2 = airy2 / np.size(data, 0)
    
    airyMax = np.max(airy2)
    airyMin = np.min(airy2)
    airyCentPerc = (0.2 * (airyMax - airyMin) + airyMin) / airyMax * 100
        
    airy2 = airy2.reshape(5, 5)
    
    
    plt.figure()
    fontSize = 20
    plt.rcParams.update({'font.size': fontSize})
    plt.rcParams['mathtext.rm'] = 'Arial'
    
    if vminmax == 'auto':
        plt.imshow(airy2, cmap='hot', interpolation='nearest')
    else:
        plt.imshow(airy2, cmap='hot', interpolation='nearest', vmin=vminmax[0], vmax=vminmax[1])
    ax = plt.gca()
    
    # Major ticks
    ax.set_xticks([])
    ax.set_yticks([])
    # Labels for major ticks
    ax.set_xticklabels([])
    ax.set_yticklabels([])
    # Minor ticks
    ax.set_xticks(np.arange(-0.5, 4.5, 1), minor=True)
    ax.set_yticks(np.arange(-0.5, 5.5, 1), minor=True)
    # Gridlines based on minor ticks
    #ax.grid(which='minor', color='w', linestyle='-', linewidth=1)
    ax.tick_params(axis=u'both', which=u'both',length=0)
    
    cbar = plt.colorbar()
    cbar.ax.tick_params(labelsize=fontSize)

    if type(show_perc) is str and show_perc=="numbers":
        for i in range(5):
            for j in range(5):
                if vminmax == 'auto':
                    perc = round(airy2[i, j] / airyMax * 100)
                else:
                    perc = round(airy2[i, j] / vminmax[1] * 100)
                c="k"
                if perc < airyCentPerc:
                    c="w"
                plt.text(j, i, '{:.1f}'.format(airy2[i, j]), ha="center", va="center", color=c, fontsize=18)    
    elif show_perc:
        for i in range(5):
            for j in range(5):
                if vminmax == 'auto':
                    perc = round(airy2[i, j] / airyMax * 100)
                else:
                    perc = round(airy2[i, j] / vminmax[1] * 100)
                c="k"
                if perc < airyCentPerc:
                    c="w"
                plt.text(j, i, '{:.0f}%'.format(perc), ha="center", va="center", color=c, fontsize=18)

    if savefig != 0:
        plt.tight_layout()
        if savefig[-3:] == 'svg':
            plt.rcParams['svg.fonttype'] = 'none'
        plt.savefig(savefig, format=savefig[-3:])

    return airy

def plot_det_dist():
    det = []
    for i in range(25):
        det.append(distance2detelements(i, 12))
    det = np.resize(det, (5, 5))
    plt.figure()
    plt.imshow(det, cmap='viridis')


# -*- coding: utf-8 -*-

import numpy as np

def tt2corr(data1, data2, macro_time=1, m=50):
    """
    Calculate the auto/cross-correlation function for log spaced tau values using fft

    Parameters
    ----------
    data1 : np.ndarray
        1D array with intensity as a function of time.
    data2 : np.ndarray
        1D array with intensity as a function of time.
        For autocorr: data2 = data1
    macro_time : double, optional
        Dwell time. The default is 1.
    m : int, optional
        Number of lag times for which to calculate G. The default is 50.

    Returns
    -------
    output : np.ndarray
        2D array with first column lag times, second column G.

    """
    G = autocorrelation_wiener_khinchin(data1, data2, macro_time=macro_time)
    lags = G[:, 0]
    
    # Generate log-spaced indices for positive lags
    positive_lags = lags[lags > 0]
    log_indices = np.logspace(0, np.log10(int(len(positive_lags) / 2 - 1)), num=int(m), dtype=int)
    log_indices = [0] + list(np.unique(log_indices))  # Ensure indices are unique
    
    # Select log-spaced lags and autocorrelation values
    log_lags = lags[log_indices]
    log_autocorr = G[log_indices, 1]
    return np.transpose([log_lags, log_autocorr])
    

def autocorrelation_wiener_khinchin(data1, data2, macro_time=1.0):
    """
    Calculate the autocorrelation function using the Wiener-Khinchin theorem.

    Parameters
    ----------
    data1 : np.ndarray
        1D array with intensity as a function of time.
    data2 : np.ndarray
        1D array with intensity as a function of time.
    macro_time : double, optional
        Dwell time. The default is 1.

    Returns
    -------
    output : np.ndarray
        2D array with first column lag times, second column G.

    """
    
    # Number of points in the signal
    n = len(data1)
    
    # Perform FFT of the signal
    
    data1_pad = np.concatenate((data1, np.zeros(len(data1)-1)))
    data2_pad = np.concatenate((data2, np.zeros(len(data2)-1)))
    
    fft1 = np.fft.fft(data1_pad)  # Zero-pad to 2n for better resolution
    fft2 = np.fft.fft(data2_pad)
    
    psd = fft1 * np.conjugate(fft2)
    
    corr = np.fft.ifft(psd).real
    g = corr[:(corr.size//2)+1]
    
    # Normalize each lag by the number of overlapping points
    n = len(g)  # Original signal length (before padding)
    overlap = np.arange(1, n + 1)  # Overlap for positive lags
    g /= overlap[::-1]  # Normalize positive lags
    g /= np.mean(data1)
    g /= np.mean(data2)
    g -= 1
    
    # Generate time lags
    n = len(g)
    tau = np.linspace(0, n-1, n) * macro_time
    
    
    #g = np.concatenate((corr[n:], corr[:n]))
    
    return np.transpose([tau, g])


import numpy as np

def time2string(seconds):
    if seconds > 3600:
        progress = int(np.floor(seconds / 3600))
        if progress > 1:
            return str(progress) + " hours"
        else:
            return str(progress) + " hour"
        
    elif seconds > 60:
        progress = int(np.floor(seconds / 60))
        if progress > 1:
            return str(progress) + " minutes"
        else:
            return str(progress) + " minute"
        
    else:
        progress = int(np.floor(seconds))
        if progress > 1 or progress == 0:
            return str(progress) + " seconds"
        else:
            return str(progress) + " second"


import numpy as np
from scipy.optimize import least_squares
import matplotlib.pyplot as plt


def fit_and_plot(y, x, fitfun, fitInfo, param, lBounds=-1, uBounds=-1, printResults=1, plot_fit=1):
    """
    Fit 1D curve with any of the implemented fit functions
    See fit_data for the meaning of the parameters
    Results are plotted
    """
    
    # number of total parameters
    Nparam = len(fitInfo)
    
    # check bounds
    if lBounds == -1:
        lBounds = np.zeros(Nparam)
    if uBounds == -1:
        uBounds = np.zeros(Nparam) + 1e4
    
    # perform fit
    fitresult = fit_data(y, x, fitfun, fitInfo, param, lBounds, uBounds)    
    
    # print results
    if printResults:
        fittedParam = fitresult.x
        for i in range(len(fittedParam)):
            print(str(fittedParam[i]))
    
    # plot results
    if plot_fit:
        plt.rcParams.update({'font.size': 20})
        plt.figure()
        plt.scatter(x, y, s=1)
        plt.plot(x, y-fitresult.fun)
    
    return fitresult


def fit_data(y, x, fitfun, fitInfo, param, lBounds, uBounds):
    """
    Fit 1D curve with any of the implemented fit functions

    Parameters
    ----------
    y : np.array()
        Vector with y values.
    x : np.array()
        Vector with x values.
    fitfun : function
        Fit function
        'exp'                 A1 * exp(-alpha1 * x) +
                            + A2 * exp(-alpha2 * x) +
                            + ...
                            + An * exp(-alphan * x) +
                            + B.
    fitInfo : np.array()
        np.array boolean vector with [A1, alpha1,..., alphan, C]
        1 if this value has to be fitted
        0 if this value is fixed during the fit.
    param : np.array
        np.array vector with start values for [A, B, C]
        A ~ 1e6*2.5e-8
        B ~ -1.05.
    lBounds : np.array
        Vector with lower bounds for all parameters.
    uBounds : np.array
        Vector with upper bounds for all parameters.

    Returns
    -------
    fitresult : fitresult
        Output of least_squares
        fitresult.x = fit results.

    """
    
    # make sure that all variables are np.arrays
    fitInfo = np.asarray(fitInfo)
    param = np.asarray(param)
    lBounds = np.asarray(lBounds)
    uBounds = np.asarray(uBounds)

    # parse fit and fixed parameters
    fitparamStart = param[fitInfo==1]
    fixedparam = param[fitInfo==0]
    lowerBounds = lBounds[fitInfo==1]
    upperBounds = uBounds[fitInfo==1]    

    # perform fit
    if fitfun == 'exp':
        fitresult = least_squares(fitfun_exp_n, fitparamStart, args=(fixedparam, fitInfo, x, y), bounds=(lowerBounds, upperBounds))

    return fitresult


def fitfun_exp_n(fitparamStart, fixedparam, fitInfo, x, y):
    """
    Exponential fit function (sum of multiple exponentials is possible)
    y = A * exp(-alpha * x) + B
    (alpha = 1 / tau with tau the lifetime)

    Parameters
    ----------
    fitparamStart : np.array
        List with starting values for the fit parameters:
        order: [A, alpha, B]
        E.g. if only A and B are fitted, this becomes a two
        element vector [1e-4, 0.2].
    fixedparam : np.array
        List with values for the fixed parameters:
        order: [A, alpha, B]
        same principle as fitparamStart.
    fitInfo : np.array
        np.array boolean vector with always 3 elements
        1 for a fitted parameter, 0 for a fixed parameter.
    x : np.array
        Vector with x values.
    y : np.array
        Vector with experimental y values.

    Returns
    -------
    res : np.array
        Residuals.

    """
    
    fitparam = np.float64(np.zeros(len(fitInfo)))
    fitparam[fitInfo==1] = fitparamStart
    fitparam[fitInfo==0] = fixedparam
    
    # calculate theoretical exponential function
    ytheo = 0
    for i in range(int((len(fitparam) - 1) / 2)):
        A = fitparam[2*i]
        alpha = fitparam[2*i+1]
        ytheo += A * np.exp(-alpha * x)
    B = fitparam[-1]
    ytheo += B
    
    # calculate residuals
    res = y - ytheo
    
    return res



import numpy as np
from scipy.optimize import least_squares
import matplotlib.pyplot as plt


"""
Fit curves to different fit functions
"""

def fit_power_law_g(G, start=0, plotIndv=True, fitInfo=np.array([1, 1, 0]), param=np.array([3.1, -1, 0])):
    # remove dwellTime field from G list
    Gkeys = list(G.__dict__.keys())
    if "dwellTime" in Gkeys:
        Gkeys.remove("dwellTime")
    
    # number of curves
    Nc = len(Gkeys)
        
    # create empty array to store fitreseults
    fitresults = np.zeros((Nc, np.sum(fitInfo)))
    
    # check how to plot
    if Nc == 25 and plotIndv == False:
        plt.rcParams.update({'font.size': 6})
        fig, axs = plt.subplots(5, 5, figsize = (9, 6))
        fig.subplots_adjust(hspace = 1, wspace=.1)
        axs = axs.ravel()
    
    for i in range(len(Gkeys)):
        GName = Gkeys[i]
        Gtemp = getattr(G, GName)
        y = Gtemp[start:,1]
        x = Gtemp[start:,0]
        fitresult = fit_curve(y, x, 'powerlaw', fitInfo, param, np.array([0, -10, -10]), np.array([100, 0, 100]), 0)
        if len(Gkeys) > 10 and 'det' in GName:
            detNr = int(GName[3:-8])
        else:
            detNr = i
        fitresults[detNr, :] = np.array(fitresult.x)
        if Nc == 25 and plotIndv == False:
            axs[detNr].plot(x, y-fitresult.fun, c='k')
            axs[detNr].scatter(x, y, s=3)
            axs[detNr].set_xscale('log')
            axs[detNr].set_title(GName)
        elif plotIndv:
            plt.rcParams.update({'font.size': 20})
            plt.figure()
            plt.scatter(x, y, s=1)
            plt.plot(x, y-fitresult.fun)
            plt.xscale('log')
            plt.title(GName)
    return fitresults


def fit_curve(y, x, fitfun, fitInfo, param, lBounds, uBounds, savefig=0):
    """
    Fit 1D curve with a power law y = A * B^x + C or other function

    Parameters
    ----------
    y : np.array()
        Vector with y values.
    x : np.array()
        Vector with x values.
    fitfun : str
        Fit function
        'powerlaw'           A * B^x + C
        'exp'                A * exp(-alpha * x) + B
        'linear'             A * x + B.
    fitInfo : np.array()
        np.array boolean vector with [A, B, C]
        1 if this value has to be fitted
        0 if this value is fixed during the fit.
    param : np.array()
        np.array vector with start values for [A, B, C]
        A ~ 1e6*2.5e-8
        B ~ -1.05.
    lBounds : np.array()
        Vector with lower bounds for the parameters.
    uBounds : np.array()
        Vector with upper bounds for the parameters.
    savefig : boolean or str, optional
        0 to not save the figure
        file name with extension to save as "png" or "eps".
        The default is 0.

    Returns
    -------
    fitresult : object
        Output of least_squares
        fitresult.x = fit results.

    """
    
    # make sure that all variables are np.arrays
    fitInfo = np.asarray(fitInfo)
    param = np.asarray(param)
    lBounds = np.asarray(lBounds)
    uBounds = np.asarray(uBounds)

    # parse fit and fixed parameters
    fitparamStart = param[fitInfo==1]
    fixedparam = param[fitInfo==0]
    lowerBounds = lBounds[fitInfo==1]
    upperBounds = uBounds[fitInfo==1]    

    # perform fit
    if fitfun == 'powerlaw':
        fitresult = least_squares(fitfun_powerlaw, fitparamStart, args=(fixedparam, fitInfo, x, y), bounds=(lowerBounds, upperBounds))
    elif fitfun == 'exp':
        fitresult = least_squares(fitfun_exp, fitparamStart, args=(fixedparam, fitInfo, x, y), bounds=(lowerBounds, upperBounds))
    elif fitfun == 'linear':
        fitresult = least_squares(fitfun_linear, fitparamStart, args=(fixedparam, fitInfo, x, y), bounds=(lowerBounds, upperBounds))
        
    #plot_fit(x, y, param, fitInfo, fitresult, savefig)

    return fitresult


def fitfun_powerlaw(fitparamStart, fixedparam, fitInfo, x, y):
    """
    Power law fit function
    y = A * B^x + C

    Parameters
    ----------
    fitparamStart : np.array()
        List with starting values for the fit parameters:
        order: [A, B, C]
        E.g. if only A and B are fitted, this becomes a two
        element vector [1e-4, 0.2].
    fixedparam : np.array()
        List with values for the fixed parameters:
        order: [A, B, C]
        same principle as fitparamStart.
    fitInfo : np.array()
        np.array boolean vector with always 3 elements
        1 for a fitted parameter, 0 for a fixed parameter.
    x : np.array()
        Vector with x values.
    y : np.array()
        Vector with experimental y values.

    Returns
    -------
    res : np.array()
        Residuals.

    """
    
    fitparam = np.float64(np.zeros(len(fitInfo)))
    fitparam[fitInfo==1] = fitparamStart
    fitparam[fitInfo==0] = fixedparam
    
    # get parameters
    A = fitparam[0]
    B = fitparam[1]
    C = fitparam[2]

    # calculate theoretical power law function
    ytheo = A * x**B + C
    
    # calcualte residuals
    res = y - ytheo
    
    return res


def fitfun_exp(fitparamStart, fixedparam, fitInfo, x, y):
    """
    Exponential fit function
    y = A * exp(-alpha * (x-dx)) + B
    (alpha = 1 / tau with tau the lifetime)

    Parameters
    ----------
    fitparamStart : np.array()
        List with starting values for the fit parameters:
        order: [A, alpha, B]
        E.g. if only A and B are fitted, this becomes a two
        element vector [1e-4, 0.2].
    fixedparam : np.array()
        List with values for the fixed parameters:
        order: [A, alpha, B]
        same principle as fitparamStart.
    fitInfo : np.array()
        np.array boolean vector with always 3 elements
        1 for a fitted parameter, 0 for a fixed parameter.
    x : np.array()
        Vector with x values.
    y : np.array()
        Vector with experimental y values.

    Returns
    -------
    res : np.array()
        Residuals.

    """
    
    fitparam = np.float64(np.zeros(len(fitInfo)))
    fitparam[fitInfo==1] = fitparamStart
    fitparam[fitInfo==0] = fixedparam
    
    # get parameters
    A = fitparam[0]
    alpha = fitparam[1]
    B = fitparam[2]
    dx = fitparam[3]

    # calculate theoretical curve
    ytheo = A * np.exp(-alpha * (x-dx)) + B
    
    # calculate residuals
    res = y - ytheo
    
    return res
   
def fitfun_linear(fitparamStart, fixedparam, fitInfo, x, y):
    """
    Linear fit function
    y = A * x + B

    Parameters
    ----------
    fitparamStart : np.array()
        List with starting values for the fit parameters:
        order: [A, B]
        E.g. if only A is fitted, this becomes a one
        element vector [1e-4].
    fixedparam : np.array()
        List with values for the fixed parameters:
        order: [A, B]
        same principle as fitparamStart.
    fitInfo : np.array()
        np.array boolean vector with always 2 elements
        1 for a fitted parameter, 0 for a fixed parameter.
    x : np.array()
        Vector with x values.
    y : np.array()
        DESCRIPTION.

    Returns
    -------
    res : np.array()
        Vector with experimental y values.

    """
    
    fitparam = np.float64(np.zeros(len(fitInfo)))
    fitparam[fitInfo==1] = fitparamStart
    fitparam[fitInfo==0] = fixedparam
    
    # get parameters
    A = fitparam[0]
    B = fitparam[1]

    # calculate theoretical curve
    ytheo = A * x + B
    
    # calculate residuals
    res = y - ytheo
    
    return res



import numpy as np
from .fcs_analytical import fcs_analytical
import copy
import multiprocessing
from joblib import Parallel, delayed

class Fitresultclass():
    pass


def mem_fit(fitparam, tau, yexp, fitfun=fcs_analytical, weights=1, startdistr='default', multiprocess=False):
    """
    Fit fcs curve with Maximum Entropy model

    Parameters
    ----------
    fitparam : list of scalars
        [n, Niter, eps2, xmu, anomD, corrVentr, shape_param] with
        n               number of diffusion components
        Niter           number of iterations
        eps2            Iteration stop criteria.
        xmu             iteration increment
        anomD           anomalous diffusion parameter
        corrVentr       weight correlation vs. entropy
        shape_param     height / width of focal volume
    tau : 1D np.array()
        tau values.
    yexp : 1D np.array()
        autocorrelation values.
    fitfun : function, optional
        fit function. The default is fcs_analytical.
    weights : 1D np.array() or scalar 1, optional
        weights for fitting. The default is 1.

    Returns
    -------
    f_fit : 1D np.array()
        fitted parameters.
    tauD : 1D np.array()
        same as input.
    G_fit : 1D np.array()
        fitted G values.

    """
    
    
    # fixedparam = [200, 10000, 5e-6, 2e-4, 1, 20]
    n_tauD      = int(fitparam[0])   # number of diffusion components
    Niter_f     = int(fitparam[1])  # number of iterations
    eps2        = fitparam[2]          # Iteration stop criteria.
    x_mu        = fitparam[3]          # iteration increment
    anomD       = fitparam[4]         # anomalous diffusion parameter
    corrVentr   = fitparam[5]     # weight correlation vs. entropy
    shape_param = fitparam[6]     # height / width of focal volume
    
    mu = copy.deepcopy(yexp)
    normc = mu[0]
    mu /= normc
    M = len(yexp)
    sigma = weights
    sigma_square = sigma**2
    
    n_tau = len(tau)
    taumin = np.log10(np.min(tau))
    taumax = np.log10(np.max(tau))
    tauD = np.logspace(taumin, taumax, n_tauD)
    
    dtauD = np.zeros(n_tauD)
    dtauD[0:n_tauD-1] = np.diff(tauD)
    dtauD[n_tauD-1] = dtauD[n_tauD-2]
    
    C0 = np.zeros((Niter_f))
    S = np.zeros((Niter_f))
    alpha = np.zeros((Niter_f))

    f_fit = startdistr
    if type(startdistr) == str:
        f_fit = np.zeros((n_tauD)) + 1 / n_tauD # start distribution
    
    fin_fit = np.zeros((n_tauD, n_tau))
    
    if multiprocess:
        Processed_list = Parallel(n_jobs=multiprocessing.cpu_count() - 1)(delayed(fitfun)(tau, 1, tauD[i], shape_param, 0, A=0, B=0, alpha=anomD) for i in list(range(n_tauD)))
        for i in range(n_tauD):
            fin_fit[i,:] = Processed_list[i] #fitfun(tau, 1, tauD[i], 3, 0, A=0, B=0, alpha=anomD)
    else:
        for i in range(n_tauD):
            fin_fit[i,:] = fitfun(tau, 1, tauD[i], shape_param, 0, A=0, B=0, alpha=anomD)
    
    G_fit = np.zeros((n_tau))
    
    iter_f = 0
    fitFound = False
    while iter_f < Niter_f and not fitFound:
        exp2 = np.sum(f_fit) # Summation of all fractions.
        f_fit = f_fit / exp2 # Normalizing fractions, so G(0)=1.
        for i in range(n_tau):
            G_fit[i] = np.sum(fin_fit[:,i] * f_fit)
        G_fit = G_fit / G_fit[0] # Normalize G_fit.
        G_fit = G_fit * np.mean(mu[1:2]) # scaled to average amplitude of first two fcs points.
        
        Diff_C_E = G_fit - mu # Residuals (G(tau)_calculated minus G(tau)_experimental).
        r = Diff_C_E / sigma # Weighted residuals.
        r_square = r**2
   
        C0[iter_f] = np.sum(r_square) / n_tau # Record weighted least-squares
        S[iter_f] = -np.sum(f_fit * np.log(f_fit)) # Entropy of each iteration.
    
        if iter_f > 1:
            if np.abs((C0[iter_f] - C0[iter_f-1]) / C0[iter_f-1]) < eps2:
                # termination at variation of chi square is less than 1e-5.
                fitFound = True
                
        # First order derivative of least-squares
        DC = np.zeros((n_tauD))
        for i in range(n_tauD):
            DC[i] = np.sum(2 * Diff_C_E * fin_fit[i,:] / sigma_square) / M
    
        # First order derivative of entropy
        DS = np.zeros((n_tauD))
        for i in range(n_tauD):
            DS[i] = -1 - np.log10(f_fit[i])
        
        # Scaling factor for balancing entropy and least squares in determining search direction.
        alpha[iter_f] = np.linalg.norm(DC) / np.linalg.norm(DS) / corrVentr
        # Search direction construct.
        e_mu = f_fit * (alpha[iter_f] * DS - DC)
        e_mu = e_mu / np.linalg.norm(e_mu)
        # Update f_fit's using search direction.
        f_fit=f_fit+ e_mu*x_mu
        for i in range(n_tauD):
            if f_fit[i]<0:
                f_fit[i] = 0.0001
            elif f_fit[i] == 0:
                f_fit[i] = 0.0001
    
        iter_f += 1
    
    return f_fit, tauD, G_fit*normc


def mem_fit_free_diffusion(param, tau, yexp, weights=1):
    if len(param) < 6:
        # add default shape parameter to parameter list in case not given as input
        param.append(3)
    [f_fit, tauD, G_fit] = mem_fit(param, tau, yexp, fitfun=fcs_analytical, weights=weights)
    # maxind = np.r_[True, f_fit[1:] > f_fit[:-1]] & np.r_[f_fit[:-1] > f_fit[1:], True]
    # Npeaks = np.sum(maxind == True)
    # print(Npeaks)
    # if Npeaks > 0:
    #     tauD = np.asarray(maxind.astype(int)) * np.asarray(f_fit) + 1e-7
    #     startdistr = tauD / np.sum(tauD)
    #     print(startdistr)
    #     [f_fit, tauD, G_fit] = mem_fit(param, tau, yexp, fitfun=fcs_analytical, weights=weights, startdistr=startdistr)
    
    fitresult = Fitresultclass()
    fitresult.x = f_fit
    fitresult.fun = yexp - G_fit
    fitresult.fun /= weights
    fitresult.tauD = tauD
    return fitresult




import numpy as np
from datetime import datetime
from .extract_spad_photon_streams import extract_spad_photon_streams
from joblib import Parallel, delayed


class Correlations:
    pass


def atimes_2_corrs(data, list_of_corr, accuracy=50, taumax="auto", perform_coarsening=True, logtau=True, split=10):
    """
    Calculate correlations between several photon streams with arrival times
    stored in macrotimes

    Parameters
    ----------
    data : object
        Object having fields det0, det1, ..., det24 which contain
            the macrotimes of the photon arrivals [in a.u.].
    list_of_corr : list
        List of correlations to calculate.
    accuracy : int, optional
        Accurracy with which to calulcate the correlation. The default is 50.
    taumax : float or string, optional
        Maximum tau value for which to calculate G. The default is "auto".
    perform_coarsening : Boolean, optional
        Apply coarsening. The default is True.
    logtau : TYPE, optional
        Use logarithmically spaced tau values. The default is True.
    split : float, optional
        Chunk size with which to split the data [s]. The default is 10.

    Returns
    -------
    G : object
        correlations.

    """
    
    if taumax == "auto":
        taumax = 1 / data.macrotime
    
    G = Correlations()
    
    Ndet = 21
    calcAv = False
    if 'av' in list_of_corr:
        # calculate the correlations of all channels and calculate average
        list_of_corr.remove('av')
        list_of_corr += list(range(Ndet))
        calcAv = True
    
    for corr in list_of_corr:
        # EXTRACT DATA
        if type(corr) == int:
            dataExtr = getattr(data, 'det' + str(corr))
            t0 = dataExtr[:, 0]
            corrname = 'det' + str(corr)
        elif corr == "sum5" or corr == "sum3":
            print("Extracting and sorting photons")
            dataExtr = extract_spad_photon_streams(data, corr)
            t0 = dataExtr[:, 0]
            corrname = corr
        
        # CALCULATE CORRELATIONS
        duration = t0[-1] * data.macrotime
        Nchunks = int(np.floor(duration / split))
        # go over all filters
        for j in range(np.shape(dataExtr)[1] - 1):
            for chunk in range(Nchunks):
                corrstring = corrname + "F" + str(j) + "_chunk" + str(chunk)
                print("Calculating correlation " + corrstring)
                tstart = chunk * split / data.macrotime
                tstop = (chunk + 1) * split / data.macrotime
                tchunk = t0[(t0 >= tstart) & (t0 < tstop)]
                tchunk -= tchunk[0]
                print("   Nphotons: " + str(len(tchunk)))
                if j == 0:
                    # no filter
                    Gtemp = atimes_2_corr(tchunk, tchunk, [1], [1], data.macrotime, accuracy, taumax, perform_coarsening, logtau)
                else:
                    # filters
                    w0 = dataExtr[:, j+1]
                    wchunk = w0[(t0 >= tstart) & (t0 < tstop)]
                    Gtemp = atimes_2_corr(tchunk, tchunk, wchunk, wchunk, data.macrotime, accuracy, taumax, perform_coarsening, logtau)
                setattr(G, corrstring, Gtemp)
            # average over all chunks
            print("   Calculating average correlation")
            listOfFields = list(G.__dict__.keys())
            listOfFields = [i for i in listOfFields if i.startswith(corrname + "F" + str(j) + "_chunk")]
            Gav = sum(getattr(G, i) for i in listOfFields) / len(listOfFields)
            setattr(G, corrname + "F" + str(j) + '_average', Gav)
    
    if calcAv:
        # calculate average correlation
        for f in range(np.shape(dataExtr)[1] - 1):
            # start with correlation of detector 20 (last one)
            Gav = getattr(G, 'det' + str(Ndet-1) + 'F' + str(f) + '_average')
            # add correlations detector elements 0-19
            for det in range(Ndet - 1):
                Gav += getattr(G, 'det' + str(det) + 'F' + str(f) + '_average')
            # divide by the number of detector elements to get the average
            Gav = Gav / Ndet
            # store average in G
            setattr(G, 'F' + str(f) + '_average', Gav)
    
    return G


def atimes_2_corr(t0, t1, w0, w1, macroTime, accuracy=50, taumax="auto", perform_coarsening=True, logtau=True):
    """
    Calculate correlation between two photon streams with arrival times t0 and t1
    Inspired by Wahl et al., Opt. Expr. 11 (26), 2003

    Parameters
    ----------
    t0 : np.array()
        Vector with arrival times channel 1 [a.u.].
    t1 : np.array()
        Vector with arrival times channel 2 [a.u.]..
    w0 : np.array()
        Vector with (filtered) weights channel 1 [a.u.].
    w1 : np.array()
        Vector with (filtered) weights channel 2 [a.u.]..
    macroTime : float
        Multiplication factor for the arrival times vectors [s].
    accuracy : int, optional
        Number of tau values for which G is calculated before delta tau is
        doubled
        E.g. accuracy = 3 yields:
            tau = [1, 2, 3, 5, 7, 9, 13, 17, 21, 29, 37, 45,...]. The default is 50.
    taumax : float or string, optional
        Maximum tau value for which to calculate the correlation
        If left empty, 1/10th of the measurement duration is used. The default is "auto".
    perform_coarsening : Boolean, optional
        Apply time trace coarsening for more efficient (and more accurate)
        correlation calculation. The default is True.
    logtau : Boolean, optional
        Use logarithmically spaced tau values. The default is True.

    Returns
    -------
    np.array()
        [N x 2] matrix with tau and G values.

    """
    
    
    # convert t0 and t1 lists to array
    t0 = np.asarray(t0)
    t1 = np.asarray(t1)
    
    # check max tau value
    if taumax=="auto":
        taumax = t0[-1] / 10
    
    # generate list [tau] with logarithmically distributed tau values
    t = 0
    tau = []
    step = 1
    if logtau:
        while t <= taumax:
            for i in range(accuracy):
                t += step
                # make sure t is multiple of step (needed for coarsing)
                t = int(np.floor(t / step) * step)
                tau.append(t)
            step *= 2
        tau = [i for i in tau if i <= taumax]
        tau = [0] + tau
    else:
        tau = np.linspace(0, taumax, taumax+1, dtype=int)
    
    # create array for g and weights
    N = len(tau)
    g = np.zeros(N)
    if len(w0) == 1:
        w0 = np.ones(len(t0))
    if len(w1) == 1:
        w1 = np.ones(len(t1))
    
    # coarse factor
    c = 1

    printDateTime = False
    if len(t0) > 100000:
        printDateTime = True
    
    # calculate g for each tau value
    if printDateTime:
        print(datetime.now())
    
    
    # tau = np.asarray(tau)
    # tau_diff = tau[1:] - tau[0:-1]
    # g = []
    
    # for tau_d in list(np.sort(np.unique(tau_diff))):
    #     idx = tau_diff == tau_d
    #     idxT = np.where(idx==True)[0]
    #     tau_list_temp = list(tau[idxT + 1])
    #     if tau_d == 1:
    #         tau_list = [0] + tau_list_temp
    #     else:
    #         tau_list = tau_list_temp
        
        
        
    #     gtemp = Parallel(n_jobs=-2)(delayed(atimes_2_corr_single)(t0, t1, w0, w1, t, c) for t in tau_list)
    #     g += gtemp
    #     # perform coarsening
    #     c *= 2
    #     [t0, w0, ind] = time_coarsening(t0, w0)
    #     [t1, w1, ind] = time_coarsening(t1, w1)
    
    # g = np.transpose(np.asarray(g))
        
    for i in range(N):
        t = tau[i]
        g[i] = atimes_2_corr_single(t0, t1, w0, w1, t/c, c)
        if perform_coarsening and np.mod(i, accuracy) == 0 and i != 0:
            # change in step size: perform coarsening
            c *= 2
            [t0, w0, ind] = time_coarsening(t0, w0)
            [t1, w1, ind] = time_coarsening(t1, w1)
        
    if printDateTime:
        print(datetime.now())
    
    tau = np.asarray(tau) * macroTime
    
    return np.transpose([tau, g])


def atimes_2_corr_single(t0, t1, w0, w1, tau, c):
    """
    Calculate single correlation value between two photon streams with arrival
    times t0 and t1, weight vectors w0 and w1, and tau value tau

    Parameters
    ----------
    t0 : np.array()
        Vector with arrival times channel 1
        [multiples of minimum coarsed macrotime].
    t1 : np.array()
        Vector with arrival times channel 2
        [multiples of minimum coarsed macrotime].
    w0 : np.array()
        Vector with weight values channel 1.
    w1 : np.array()
        Vector with weight values channel 2.
    tau : np.array()
        tau value for which to calculate the correlation
        [multiples of minimum coarsed macrotime].
    c : int
        Coarsening factor (power of 2).

    Returns
    -------
    g : float
        Single value g(tau).

    """
    
    # calculate time shifted vector
    t1 = t1 + tau
    
    # find intersection between t0 and t1
    [tauDouble, idxt0, idxt1] = np.intersect1d(t0, t1, return_indices=True)
    
    # overlap time
    T = (t0[-1] - tau + 1) * c
    
    # calculate autocorrelation value
    G = np.sum(w0[idxt0] * w1[idxt1]) / T / c
    
    # normalize G
    I0 = np.sum(w0[t0 >= tau]) / T
    I1 = np.sum(w1[t1<=t0[-1]]) / T
    g = G / I0 / I1 - 1
    
    return g


def time_coarsening(t, w):
    # divide all arrival times by 2
    t = np.floor(t / 2)
    
    # check for duplicate values
    ind = np.where(np.diff(t) == 0)[0]
    
    # sum weights
    w[ind] += w[ind+1]
    
    # delete duplicates
    w = np.delete(w, ind+1)
    t = np.delete(t, ind+1)
    
    return [t, w, ind]
    


import numpy as np


def moving_average(data, n=3):
    """
    calculate moving average from list of values

    Parameters
    ----------
    data : np.array()
        np.array with data values.
    n : int, optional
        width of the window. The default is 3.

    Returns
    -------
    np.array()
        np.array with data values with moving average applied.

    """
    
    dataCum = np.cumsum(data, dtype=float)
    dataCum[n:] = dataCum[n:] - dataCum[:-n]
    
    return dataCum[n-1:] / n



import numpy as np
import matplotlib.pyplot as plt


def filter_ap(ks, tau, T, b, norm_exp=False, plot_fig=True):
    """
    Use fluorescence lifetime to generate filters that remove afterpulsing
    from fcs data, based on ideal exponential decay
    Based on Enderlein and Gregor, Rev. Sci. Instr., 76, 2005 and
    Kapusta et al., J. Fluor., 17, 2007.
    
    Function call examples:
        F = filter_ap(0, 23.78, T, 0.028)

    Parameters
    ----------
    ks : scalar
        Parameter describing STED decay rate (ks = 0 for no STED).
    tau : scalar
        Fluorescence lifetime from exponential fit of the histogram.
    T : scalar
        Number of data points in time.
    b : scalar
        Background value from exponential fit of the histogram.
    norm_exp : boolean, optional
        Normalize the histogram of the exponential decay by setting
        the sum of the bins to 1. For a full decay, this is always
        the case, since the integral of (1/tau * exp(-t/tau)) running
        from 0 to inf is 1.. The default is False.
    plot_fig : boolean, optional
        Plotting. The default is True.

    Returns
    -------
    F : 2D numpy array
        2D array with 2 rows, one row for each filter function.

    """

    t = np.linspace(0, T-1, int(T))

    # histogram 1 with the fluorescence decay
    decayNorm = (1/tau) * np.exp(-t/tau) / (1 + 0.5 * ks * t / tau)
    if norm_exp:
        decayNorm = decayNorm / np.sum(decayNorm)
    
    # histogram 2 with the background, always normalized
    decayNorm2 = np.ones((1,T)) / T
    
    # concatenate both row vectors vertically
    decayNorm = np.concatenate((np.transpose(decayNorm[:,np.newaxis]), decayNorm2), axis=0)
    
    # total average decay as expected from theory
    ItheorTOT = b + np.exp(-t/tau) / (1 + 0.5 * ks * t / tau)
    
    # matrix D as in Enderlein paper
    D = np.diag(1 / ItheorTOT)
    
    # filter formula
    F = np.matmul(np.matmul(decayNorm, D), np.transpose(decayNorm))
    F = np.linalg.inv(F)
    F = np.matmul(np.matmul(F, decayNorm), D)
    
    if plot_fig:
        plt.figure()
        plt.plot(F[0,:])
        plt.plot(F[1,:])
    
    return F


def filter_2c(hist1, hist2, b, plot_fig=True):
    """
    Use histogram information to generate filters to split the data in two
    time traces
    Function call examples:
        F = filter_ap(histFluor, np.array(ones(500)))

    Parameters
    ----------
    hist1 : 1D numpy array
        Vector with the histogram values for component 1
        (can be theoretical or experimental curve
         can - but does not have to - be normalized.
    hist2 : 1D numpy array
        Vector with the histogram values for component 2
        (can be theoretical or experimental curve
         can - but does not have to - be normalized..
    b : TYPE
        Number, scaling factor describing the weight of hist2
        with respect to hist1
        equal to amplitude(hist2) / amplitude(hist1) for unnormalized
        histograms.
    plot_fig : boolean, optional
        Plot figure. The default is True.

    Returns
    -------
    F : 2D numpy array
        2D array with 2 rows, one row for each filter function.

    """
    
    # normalize histograms
    hist1 = hist1 / np.sum(hist1)
    hist2 = hist2 / np.sum(hist2)
    
    # concatenate both row vectors vertically
    histAll = np.concatenate((np.transpose(hist1[:,np.newaxis]), np.transpose(hist2[:,np.newaxis])), axis=0)
    
    # total histogram
    histTot = hist1 / np.max(hist1) + b * hist2 / np.max(hist2)
    
    # matrix D as in Enderlein paper
    D = np.diag(1 / histTot)
    
    # filter formula
    F = np.matmul(np.matmul(histAll, D), np.transpose(histAll))
    F = np.linalg.inv(F)
    F = np.matmul(np.matmul(F, histAll), D)
    
    if plot_fig:
        plt.figure()
        plt.plot(F[0,:])
        plt.plot(F[1,:])
    
    return F


import numpy as np
from scipy.optimize import least_squares


def fit_gauss_2d(y, fitInfo, param, weights=1):
    """
    Fit 2D Gaussian function for STICS analysis

    Parameters
    ----------
    y : np.array
        Matrix with function values.
    fitInfo : np.array
        np.array boolean vector with always 5 elements
        1 for a fitted parameter, 0 for a fixed parameter
        order: [x0, y0, A, sigma, offset]
        E.g. to fit sigma and offset this becomes [0, 0, 0, 1, 1].
    param : np.array
        List with starting values for all parameters:
        order: [x0, y0, A, sigma, offset].
    weights : np.array, optional
        Matrix with weight values 1 for unweighted fit. The default is 1.

    Returns
    -------
    fitresult : np.array
        Result of the nonlinear least squares fit.

    """
    
    fitInfo = np.array(fitInfo)
    param = np.array(param)
    lowerBounds = np.array([0, 0, 0, 1e-10, -100])
    upperBounds = np.array([1e6, 1e6, 1e6, 1e6, 100])
    
    fitparamStart = param[fitInfo==1]
    fixedparam = param[fitInfo==0]
    lowerBounds = lowerBounds[fitInfo==1]
    upperBounds = upperBounds[fitInfo==1]
    
    fitresult = least_squares(fitfun_gauss2d, fitparamStart, args=(fixedparam, fitInfo, y, weights), bounds=(lowerBounds, upperBounds))
    
    return fitresult
    

def fitfun_gauss2d(fitparamStart, fixedparam, fitInfo, y, weights=1):
    """
    Calculate residuals 2D Gaussian function for STICS analysis

    Parameters
    ----------
    fitparamStart : np.array
        List with starting values for the fit parameters:
        order: [x0, y0, A, sigma, offset]
        with    x0, y0      position of the Gaussian center
                A           amplitude of the Gaussian
                sigma       Standard deviation of the Gaussian
                            w0 = 2 * sigma
                            with w0 1/e^2 radius
                offset      dc offset.
    fixedparam : np.array
        List with values for the fixed parameters.
    fitInfo : np.array
        Boolean list of always 5 elements
        1   fit this parameters
        0   keep this parameter fixed.
    y : np.array
        Matrix with experimental function values.
    weights : np.array, optional
        Matrix with fit weights 1 for unweighted fit. The default is 1.

    Returns
    -------
    res : np.array
        Vector with residuals: f(param) - y.

    """
    
    param = np.float64(np.zeros(5))
    param[fitInfo==1] = fitparamStart
    param[fitInfo==0] = fixedparam
    
    x0 = param[0]
    y0 = param[1]
    A = param[2]
    sigma = param[3]
    offset = param[4]
    Ny = np.shape(y)[0]
    Nx = np.shape(y)[1]
    
    res = np.reshape(weights * (normal2d(Nx, Ny, x0, y0, A, sigma, offset) - y), [Nx*Ny, 1])
    res = np.ravel(res)

    return res


def normal2d(Nx, Ny, x0, y0, A, sigma, offset):
    """
    Calculate normal 2D distribution

    Parameters
    ----------
    Nx : int
        Number of columns of the data matrix.
    Ny : int
        Number of rows of the data matrix.
    x0 : int
        Column and row number of the center of the Gaussian function.
    y0 : int
        Column and row number of the center of the Gaussian function.
    A : float
        Amplitude.
    sigma : float
        Standard deviation of the Gaussian function.
    offset : float
        DC component.

    Returns
    -------
    out : 2D np.array()
        Matrix with normal distribution
        y = exp(- ((x-x0)^2 + (y-y0)^2) / (2*sigma^2)).

    """
    
    meshgrid = np.meshgrid(np.linspace(0, Nx-1, Nx), np.linspace(0, Ny-1, Ny))
    x = meshgrid[0]
    y = meshgrid[1]
    out = A * np.exp(-((x-x0)**2 + (y-y0)**2) / 2 / sigma**2) + offset
    return out


"""
    Stored in memory:
    root.image
    root.imageRaw
    root.firstFile
    root.activeFileButton
    root.FFSsettings
    (root.defaultMetaData
    root.airyDummy
    root.xtrace
    root.ytrace)
    
    To do: save corrsettings average?"""

import os
import shutil
import numpy as np
import h5py
from PyQt5.QtWidgets import QFileDialog
from datetime import datetime

from .analysis_settings import FFSlib, FFSfile, FFSmetadata, FFScorr, CorrSettings, FitSingleObj, CorrFit
from ..fcs.fcs2corr import Correlations
from .fitmodel_class import FitModel

from ..tools.csv2array import csv2array, array2csv
from ..tools.checkfname import checkfname

#current = os.path.dirname(os.path.realpath('restore_session.py'))
#parent = os.path.dirname(current)
#sys.path.append(parent)


def restorelib(libfile, root=0):
    # restore session from .ffs or .ffz lib file
    if str(libfile).endswith('.ffs'):
        lib = restorelib_ffs(libfile, root)
        return lib
    if str(libfile).endswith('.ffz') or str(libfile).endswith('.zip'):
        lib = restorelib_ffz(libfile, root)
        return lib


def restorelib_ffz(libfile, root=0):
    # unzip file
    if root != 0:
        root.progress = 0
        root.progressMessage = 'Opening file...'
    folder = unzip(libfile)
    
    # create session library
    lib = FFSlib()
    lib.version = read_text(os.path.join(folder, 'version.txt'))
    lib.notes = read_text(os.path.join(folder, 'notes.txt'))
    lib.active_image = read_text(os.path.join(folder, 'activeImage.txt'), dtype=int)
    if lib.version != '1.0':
        lib.date_created = read_text(os.path.join(folder, 'date_created.txt'), dtype=str)
        lib.date_modified = read_text(os.path.join(folder, 'date_modified.txt'), dtype=str)
    
    # walk through folders
    images = next(os.walk(folder))[1]
    Nimages = len(images)
    for i in range(Nimages):
        if root != 0:
            root.progress = i / Nimages
            root.progressMessage = 'Reading image ' + str(i+1) + '...'
        imagefolder = os.path.join(folder, 'image' + str(i))
        imagepath = os.path.join(imagefolder, 'image.csv')
        image = csv2array(imagepath)
        imagenamepath = os.path.join(imagefolder, 'imageName.txt')
        fname = read_text(imagenamepath)
        # add image
        lib.add_image(image, fname)
        imageObj = lib.get_image(i)
        # active file
        imageObj.active_ffs = read_text(os.path.join(imagefolder, 'activeFFS.txt'), dtype=int)
        
        # check for files in image
        files = next(os.walk(imagefolder))[1]
        Nfiles = len(files)
        for j in range(Nfiles):
            if root != 0:
                root.progress = i / Nimages + j / (Nfiles * Nimages)
                root.progressMessage = 'Reading FFS file ' + str(j+1) + ', image ' + str(i+1) + '...'
            fileObj = FFSfile()
            filefolder = os.path.join(imagefolder, 'file' + str(j))
            fileObj.fname = read_text(os.path.join(filefolder, 'fname.txt'))
            fileObj.label = read_text(os.path.join(filefolder, 'label.txt'))
            fileObj.timetrace = csv2array(os.path.join(filefolder, 'timetrace.csv'))
            if -1 in fileObj.timetrace:
                fileObj.timetrace = None
            fileObj.airy = csv2array(os.path.join(filefolder, 'airy.csv'))
            if -1 in fileObj.airy:
                fileObj.airy = None
            fileObj.active_analysis = read_text(os.path.join(filefolder, 'activeAnalysis.txt'), dtype=int)
            md = FFSmetadata()
            md.num_pixels = read_text(os.path.join(filefolder, 'numberOfPixels.txt'), dtype=int)
            md.num_lines = read_text(os.path.join(filefolder, 'numberOfLines.txt'), dtype=int)
            md.num_frames = read_text(os.path.join(filefolder, 'numberOfFrames.txt'), dtype=int)
            md.range_x = read_text(os.path.join(filefolder, 'rangeX.txt'), dtype=int)
            md.range_y = read_text(os.path.join(filefolder, 'rangeY.txt'), dtype=int)
            md.range_z = read_text(os.path.join(filefolder, 'rangeZ.txt'), dtype=int)
            md.num_datapoints = read_text(os.path.join(filefolder, 'numberOfDataPoints.txt'), dtype=int)
            md.hold_off_x5 = read_text(os.path.join(filefolder, 'holdOffx5.txt'), dtype=int)
            md.hold_off = read_text(os.path.join(filefolder, 'holdOff.txt'), dtype=int)
            md.time_resolution = read_text(os.path.join(filefolder, 'timeResolution.txt'), dtype=float)
            md.dwelltime = read_text(os.path.join(filefolder, 'dwellTime.txt'), dtype=float)
            md.duration = read_text(os.path.join(filefolder, 'duration.txt'), dtype=float)
            md.pxsize = read_text(os.path.join(filefolder, 'pxsize.txt'), dtype=float)
            yc = 0
            xc = 0
            coords = check_none(csv2array(os.path.join(filefolder, 'coords.csv')))
            
            if coords is not None:
                yc = int(coords[0])
                xc = int(coords[1])
            md.coords = [yc, xc]
            fileObj.metadata = md
            imageObj.ffs_list.append(fileObj)
            # check for analysis in file
            fileObj = imageObj.get_ffs_file(j)
            analyses = next(os.walk(filefolder))[1]
            Nanalyses = len(analyses)
            for k in range(Nanalyses):
                if root != 0:
                    root.progressMessage = 'Reading correlation analysis ' + str(k+1) + ', FFS file ' + str(j+1) + ', image ' + str(i+1) + '...'
                analysisfolder = os.path.join(filefolder, 'analysis' + str(k))
                
                corrObj = FFScorr()
                
                corrObj.mode = read_text(os.path.join(analysisfolder, 'mode.txt'))
                corrObj.active_fit = read_text(os.path.join(analysisfolder, 'activeFit.txt'), dtype=int)
                corrSett = CorrSettings()
                
                elString = read_text(os.path.join(analysisfolder, 'elements.txt'))
                elList = elString.split(", ")
                elList[0]=elList[0][1:]
                elList[-1]=elList[-1][0:-1]
                elListOut = []
                for l in range(len(elList)):
                    if elList[l][0] == "'":
                        elListOut.append(elList[l][1:-1])
                    else:
                        elListOut.append(int(elList[l]))
                corrSett.elements = elListOut
                
                gString = read_text(os.path.join(analysisfolder, 'listOfG.txt'))
                gList = gString.split(", ")
                gList[0]=gList[0][1:]
                gList[-1]=gList[-1][0:-1]
                gListOut = []
                for l in range(len(gList)):
                    if gList[l][0] == "'":
                        gListOut.append(gList[l][1:-1])
                    else:
                        gListOut.append(int(gList[l]))
                corrSett.list_of_g = gListOut
                
                corrSett.average = read_text(os.path.join(analysisfolder, 'average.txt'), dtype=str)
                corrSett.algorithm = read_text(os.path.join(analysisfolder, 'algorithm.txt'), dtype=str)
                corrSett.resolution = read_text(os.path.join(analysisfolder, 'resolution.txt'), dtype=int)
                corrSett.chunksize = read_text(os.path.join(analysisfolder, 'chunksize.txt'), dtype=float)
                corrSett.chunks_off = [int(l) for l in csv2array(os.path.join(analysisfolder, 'chunksOff.csv'))]
                corrObj.settings = corrSett
                # get all .csv files with correlations
                list_filesAll = os.listdir(analysisfolder)
                corrFiles = []
                for l in list_filesAll:
                    if l == 'chunksOff.csv' or l =='fits' or l[-4:] == '.txt':
                        pass
                    else:
                        corrFiles.append(l)
                corrs = Correlations()
                for l in corrFiles:
                    setattr(corrs, l[0:-4], csv2array(os.path.join(analysisfolder, l)))
                corrs.dwellTime = -1
                if len(corrFiles) > 0:
                    corrObj.corrs = corrs
                fileObj.analysis_list.append(corrObj)
                # check overall fit
                anObj = fileObj.get_analysis(k)
                fitfolder = os.path.join(analysisfolder, 'fits')
                fits = next(os.walk(fitfolder))[1]
                Nfits = len(fits)
                for l in range(Nfits):
                    cfitObj = CorrFit()
                    # check individual fits
                    fitsinglefolder = os.path.join(fitfolder, 'fit' + str(l))
                    singlefits = next(os.walk(fitsinglefolder))[1]
                    Nsinglefits = len(singlefits)
                    for m in range(Nsinglefits):
                        fitindvfolder = os.path.join(fitsinglefolder, 'fitSingleCurve' + str(m))
                        fdata = read_text(os.path.join(fitindvfolder, 'data.txt'))
                        ffitfunctionLabel = read_text(os.path.join(fitindvfolder, 'fitfunctionLabel.txt'))
                        ffitarray = np.array([int(i) for i in csv2array(os.path.join(fitindvfolder, 'fitarray.csv'))])
                        fstartvalues = csv2array(os.path.join(fitindvfolder, 'startvalues.csv'))
                        fparamFactors10 = csv2array(os.path.join(fitindvfolder, 'paramFactors10.csv'))
                        fitr = csv2array(os.path.join(fitindvfolder, 'fitrange.csv'))
                        fitstart = int(fitr[0])
                        fitstop = int(fitr[1])
                        ffitrange = [fitstart, fitstop]
                        indfitObj = FitSingleObj(fdata, FitModel(), ffitarray, fstartvalues, ffitrange)
                        indfitObj.data = fdata
                        indfitObj.fitfunction_label = ffitfunctionLabel
                        indfitObj.fitarray = ffitarray
                        indfitObj.startvalues = fstartvalues
                        indfitObj.param_factors10 = fparamFactors10
                        indfitObj.fitrange = ffitrange
                        indfitObj.D = read_text(os.path.join(fitindvfolder, 'D.txt'), float)
                        #indfitObj.fitfunction = read_text(os.path.join(fitindvfolder, 'fitfunction.txt'))
                        indfitObj.fitresult = csv2array(os.path.join(fitindvfolder, 'fitresult.csv'))
                        indfitObj.minbound = csv2array(os.path.join(fitindvfolder, 'minbound.csv'))
                        indfitObj.maxbound = csv2array(os.path.join(fitindvfolder, 'maxbound.csv'))
                        indfitObj.paramidx = [int(i) for i in csv2array(os.path.join(fitindvfolder, 'paramidx.csv'))]
                        
                        indfitObj.w0 = read_text(os.path.join(fitindvfolder, 'w0.txt'), float)
                        cfitObj.fit_all_curves.append(indfitObj)
                    anObj.fits.append(cfitObj)
    if root != 0:
        root.progressMessage = 'Almost there...'
    shutil.rmtree(folder)
    return lib

def restorelib_ffs(libfile, root=0):
    # restore session from .ffs lib file (h5)
    # unzip file
    if root != 0:
        root.progress = 0
        root.progressMessage = 'Opening file...'
    
    # create session library
    lib = FFSlib()
    with h5py.File(libfile) as f:
        lib.version = check_none(f['info'].attrs['version'])
        lib.notes = check_none(f['info'].attrs['notes'])
        
        active_im_str = 'active_image'
        if lib.version == '1.0':
            active_im_str = 'activeImage'
        lib.active_image = check_none(f['info'].attrs[active_im_str], dtype=int)
        
        if lib.version != '1.0':
            lib.date_created = check_none(f['info'].attrs['date_created'], dtype=str)
            lib.date_modified = check_none(f['info'].attrs['date_modified'], dtype=str)
        
        # walk through images
        imgList = [img for img in list(f.keys()) if img[0:5]=='image']
        Nimages = len(imgList)
        
        for i in range(Nimages):
            if root != 0:
                root.progress = i / Nimages
                root.progressMessage = 'Reading image ' + str(i+1) + '...'
            
            image = f['image' + str(i) + '/image'][:]
            
            im_name_str = 'image_name'
            if lib.version == '1.0':
                im_name_str = 'imageName'
            fname = f['image' + str(i)].attrs[im_name_str]
            
            # add image
            lib.add_image(image, fname)
            imageObj = lib.get_image(i)
            
            # active file
            active_file_str = 'active_ffs'
            if lib.version == '1.0':
                active_file_str = 'activeFFS'
            imageObj.active_ffs = check_none(f['image' + str(i)].attrs[active_file_str], dtype=int)
            
            # check for files in image
            fileList = list(f['image' + str(i)].keys())
            files = [file for file in fileList if file[0:4] == 'file']
            Nfiles = len(files)
            for j in range(Nfiles):
                currFile = 'image' + str(i) + '/file' + str(j)
                if root != 0:
                    root.progress = i / Nimages + j / (Nfiles * Nimages)
                    root.progressMessage = 'Reading FFS file ' + str(j+1) + ', image ' + str(i+1) + '...'
                fileObj = FFSfile()
                
                args_str_new = ['fname', 'label', 'active_analysis']
                args_str_old = args_str_new
                if lib.version == '1.0':
                    args_str_old = ['fname', 'label', 'activeAnalysis']
                for k in range(len(args_str_new)):
                    setattr(fileObj, args_str_new[k], check_none(f[currFile + '/'].attrs[args_str_old[k]]))
                
                for arg in ['timetrace', 'airy']:
                    dummy = check_none(f[currFile + '/' + arg])
                    if dummy is not None:
                        dummy = np.array(dummy)
                    setattr(fileObj, arg, dummy)
                
                if -1 in fileObj.timetrace:
                    fileObj.timetrace = None
                
                if -1 in fileObj.airy:
                    fileObj.airy = None
                
                md = FFSmetadata()
                
                args_str_new = ['num_pixels', 'num_lines', 'num_frames', 'range_x', 'range_y', 'range_z', 'num_datapoints', 'hold_off_x5', 'hold_off']
                args_str_old = args_str_new
                if lib.version == '1.0':
                    args_str_old = ['numberOfPixels', 'numberOfLines', 'numberOfFrames', 'rangeX', 'rangeY', 'rangeZ', 'numberOfDataPoints', 'holdOffx5', 'holdOff']
                for k in range(len(args_str_new)):
                    setattr(md, args_str_new[k], check_none(f[currFile].attrs[args_str_old[k]], dtype=int))
                
                args_str_new = ['time_resolution', 'dwelltime', 'duration', 'pxsize']
                args_str_old = args_str_new
                if lib.version == '1.0':
                    args_str_old = ['timeResolution', 'dwellTime', 'duration', 'pxsize']                
                for k in range(len(args_str_new)):
                    setattr(md, args_str_new[k], check_none(f[currFile].attrs[args_str_old[k]], dtype=float))
                
                coords = check_none(f[currFile + '/coords'][:])
                if coords is not None:
                    coords = coords.astype(int)
                md.coords = coords
                fileObj.metadata = md
                imageObj.ffs_list.append(fileObj)
                
                # check for analysis in file
                fileObj = imageObj.get_ffs_file(j)
                
                analList = list(f[currFile].keys())
                analyses = [anal for anal in analList if anal[0:8] == 'analysis']
                Nanalyses = len(analyses)
                for k in range(Nanalyses):
                    if root != 0:
                        root.progressMessage = 'Reading correlation analysis ' + str(k+1) + ', FFS file ' + str(j+1) + ', image ' + str(i+1) + '...'
                    currAnal = currFile + '/analysis' + str(k)
                    
                    corrObj = FFScorr()
                    corrObj.mode = check_none(f[currAnal].attrs['mode'], dtype=str)
                    
                    active_fit_str = 'activeFit' if lib.version == '1.0' else 'active_fit'
                    corrObj.active_fit = check_none(f[currAnal].attrs[active_fit_str], dtype=int)
                    corrSett = CorrSettings()
                    
                    elList = check_none(f[currAnal].attrs['elements'])
                    elListOut = []
                    for l in range(len(elList)):
                        elListOut.append(elList[l])
                    
                    glist_str = 'listOfG' if lib.version == '1.0' else 'list_of_g'
                    gList = check_none(f[currAnal].attrs[glist_str])
                   
                    try:
                        gListOut = []
                        for l in range(len(gList)):
                            gListOut.append(gList[l])
                    except:
                        gListOut = None
                    
                    
                    corrSett.elements = elListOut
                    corrSett.list_of_g = gListOut
                    corrSett.average = getattr(f[currAnal].attrs, 'average', None)
                    corrSett.algorithm = f[currAnal].attrs['algorithm']
                    corrSett.resolution = check_none(f[currAnal].attrs['resolution'], dtype=int)
                    corrSett.chunksize = check_none(f[currAnal].attrs['chunksize'], dtype=float)
                    
                    chunksoff_str = 'chunksOff' if lib.version == '1.0' else 'chunks_off'
                    corrSett.chunks_off = check_none(f[currAnal][chunksoff_str][:], dtype=None)
                    
                    corrObj.settings = corrSett
                    
                    Ncorrs = 0
                    corrs = Correlations()
                    corrs.dwellTime = -1 # not used
                    for corr in f[currAnal].keys():
                        if corr not in ['fits']:
                            Ncorrs += 1
                            setattr(corrs, corr, f[currAnal + '/' + corr][:])
                    
                    if Ncorrs > 1:
                        # if only 1 element found, this element is chunksoff and not used because already stored elsewhere
                        corrObj.corrs = corrs
                    
                    fileObj.analysis_list.append(corrObj)
                    
                    # check overall fit
                    anObj = fileObj.get_analysis(k)
                    fitFolder = 'image' + str(i) + '/file' + str(j) + '/analysis' + str(k) + '/fits'
                    
                    fitList = list(f[fitFolder].keys())
                    Nfits = len(fitList)
                    
                    for l in range(Nfits):
                        cfitObj = CorrFit()
                        
                        if root != 0:
                            root.progressMessage = 'Loading fit ' + str(l+1) + ', analysis' + str(k+1) +  ', FFS file ' + str(j+1) + ', image ' + str(i+1) + '...'
                    
                        fitAllCurves5 = fitFolder + '/fit' + str(l)
                        
                        Nfits3 = len(f[fitAllCurves5].keys())
                        for m in range(Nfits3):
                            if root != 0:
                                root.progressMessage = 'Loading fit--- ' + str(m) +  ', FFS file ' + str(j+1) + ', image ' + str(i+1) + '...'
                            
                            fitsingle_str = '/fitSingleCurve' if lib.version == '1.0' else '/fit_single_curve'
                            fitSingleCurveFolder = fitAllCurves5 + fitsingle_str + str(m)
                            fitSingleCurve = f[fitSingleCurveFolder]
                            
                            ffitarray = fitSingleCurve['fitarray'][:]
                            ffitresult = fitSingleCurve['fitresult'][:]
                            fmaxbound = fitSingleCurve['maxbound'][:]
                            fminbound = fitSingleCurve['minbound'][:]
                            fparamidx = fitSingleCurve['paramidx'][:]
                            fstartvalues = fitSingleCurve['startvalues'][:]
                            
                            fparam_str = 'paramFactors10' if lib.version == '1.0' else 'param_factors10'
                            fparamFactors10 = fitSingleCurve[fparam_str][:]
                            
                            fitf_str = 'fitfunctionLabel' if lib.version == '1.0' else 'fitfunction_label'
                            ffitfunctionLabel = f[fitSingleCurveFolder].attrs[fitf_str]
                            
                            ffitrange = f[fitSingleCurveFolder].attrs['fitrange']
                            fD = f[fitSingleCurveFolder].attrs['D']
                            fw0 = check_none(f[fitSingleCurveFolder].attrs['w0'])
                            fdata = f[fitSingleCurveFolder].attrs['data']
                            
                            indfitObj = FitSingleObj(fdata, FitModel(), ffitarray, fstartvalues, ffitrange)
                            indfitObj.data = fdata
                            indfitObj.fitfunction_label = ffitfunctionLabel
                            indfitObj.fitarray = ffitarray
                            indfitObj.startvalues = fstartvalues
                            indfitObj.param_factors10 = fparamFactors10
                            indfitObj.fitrange = ffitrange
                            indfitObj.D = fD
                            indfitObj.fitresult = ffitresult
                            indfitObj.minbound = fminbound
                            indfitObj.maxbound = fmaxbound
                            indfitObj.paramidx = fparamidx
                            indfitObj.w0 = fw0
                            
                            cfitObj.fit_all_curves.append(indfitObj)
                          
                            
                        anObj.fits.append(cfitObj)
    if root != 0:
        root.progressMessage = 'Almost there...'
    
    return lib

def savelib(FFSlib, root=0, fname=''):
    if fname == '':
        fname = save_ffs()
    # save session in .ffs or .ffz lib file
    if str(fname).endswith('.ffz'):
        savelib_ffz(FFSlib, root, fname)
    else:
        # make sure the file ends with .ffs
        fname = checkfname(fname, 'ffs')
        savelib_ffs(FFSlib, root, fname)
    return fname

def savelib_ffz(FFSlib, root=0, fname=''):
    if fname == '':
        fname = save_ffs()
    if fname is not None:
        if root != 0:
            root.progress = 0
            root.progressMessage = 'Saving file...'
        lib = FFSlib.lib
        # make folder to store data temporarilyy
        randString = str(int(np.random.rand()*100000))
        folderName = "Facts_tempData_" + randString
        create_dir(folderName)
        save_text("1.1", folderName + '/version.txt')
        save_text(FFSlib.notes, folderName + '/notes.txt')
        save_text(FFSlib.active_image, folderName + '/activeImage.txt')
        save_text(FFSlib.date_created, folderName + '/date_created.txt')
        save_text(str(datetime.now()), folderName + '/date_modified.txt')
        Nimages = len(lib)
        for i in range(Nimages):
            if root != 0:
                root.progress = i / Nimages
                root.progressMessage = 'Saving image ' + str(i+1) + '...'
            im = lib[i]
            # image found
            imageFolder = folderName + "/image" + str(i)
            create_dir(imageFolder)
            # store image name
            save_text(im.image_name, imageFolder + '/imageName.txt')
            # store image
            array2csv(im.image, imageFolder + '/image.csv', dtype=int)
            array2csv(im.image, imageFolder + '/imageRaw.csv', dtype=int)
            # store active file
            save_text(im.active_ffs, imageFolder + '/activeFFS.txt')
            # to through ffs files
            Nfiles = len(im.ffs_list)
            for j in range(Nfiles):
                if root != 0:
                    root.progress = i / Nimages + j / (Nimages * Nfiles)
                    root.progressMessage = 'Saving FFS file ' + str(j+1) + ', image ' + str(i+1) + '...'
                file = im.ffs_list[j]
                fileFolder = imageFolder + "/file" + str(j)
                create_dir(fileFolder)
                # save file name
                save_text(file.fname, fileFolder + '/fname.txt')
                # save label
                save_text(file.label, fileFolder + '/label.txt')
                # save time trace
                tt = file.timetrace
                if tt is None:
                    tt = np.zeros((1)) - 1
                array2csv(tt, fileFolder + '/timetrace.csv')
                # save finger print
                fp = file.airy
                if fp is None:
                    fp = np.zeros((1)) - 1
                array2csv(fp, fileFolder + '/airy.csv')
                # save active analysis
                save_text(file.active_analysis, fileFolder + '/activeAnalysis.txt')
                # save metadata
                md = file.metadata
                for key in list(md.__dict__.keys()):
                    if key != 'coords':
                        save_text(check_none(getattr(md, key)), fileFolder + '/' + key + '.txt')
                    else:
                        array2csv(check_none(getattr(md, key)), fileFolder + '/' + key + '.csv')
                # go through each analysis
                Nanalysis = len(file.analysis_list)
                for k in range(Nanalysis):
                    if root != 0:
                        root.progressMessage = 'Saving analysis ' + str(k) +  ', FFS file ' + str(j+1) + ', image ' + str(i+1) + '...'
                    analysis = file.analysis_list[k]
                    analysisFolder = fileFolder + "/analysis" + str(k)
                    create_dir(analysisFolder)
                    # save analysis mode
                    save_text(analysis.mode, analysisFolder + '/mode.txt')
                    # save elements
                    save_text(analysis.settings.elements, analysisFolder + '/elements.txt')
                    # save g list
                    save_text(analysis.settings.list_of_g, analysisFolder + '/listOfG.txt')
                    # save average list
                    save_text(analysis.settings.average, analysisFolder + '/average.txt')
                    # save algorithm
                    save_text(analysis.settings.algorithm, analysisFolder + '/algorithm.txt')
                    # save resolution
                    save_text(analysis.settings.resolution, analysisFolder + '/resolution.txt')
                    # save chunck size
                    save_text(analysis.settings.chunksize, analysisFolder + '/chunksize.txt')
                    # save chunks off
                    array2csv(analysis.settings.chunks_off, analysisFolder + '/chunksOff.csv')
                    # save active fit
                    save_text(analysis.active_fit, analysisFolder + '/activeFit.txt')
                    # save correlations
                    c = analysis.corrs
                    if c is not None:
                        for key in list(c.__dict__.keys()):
                            if key != "dwellTime":
                                # DWELLTIME IS ALREADY STORED IN METADATA
                                array2csv(getattr(c, key), analysisFolder + '/' + key + '.csv')
                    # save fits
                    fitsFolder = analysisFolder + "/fits"
                    create_dir(fitsFolder)
                    fits = analysis.fits
                    Nfits = len(fits)
                    for l in range(Nfits):
                        fitAllCurvesFolder = fitsFolder + "/fit" + str(l)
                        create_dir(fitAllCurvesFolder)
                        fitAllCurves = fits[l].fit_all_curves # contains 3 fits: central, sum3, sum5
                        Nfits3 = len(fitAllCurves)
                        for m in range(Nfits3):
                            fitSingleCurveFolder = fitAllCurvesFolder + "/fitSingleCurve" + str(m)
                            create_dir(fitSingleCurveFolder)
                            fitSingleCurve = fitAllCurves[m]
                            # save all fields
                            array2csv(fitSingleCurve.minbound, fitSingleCurveFolder + '/minbound.csv')
                            array2csv(fitSingleCurve.maxbound, fitSingleCurveFolder + '/maxbound.csv')
                            save_text(fitSingleCurve.data, fitSingleCurveFolder + '/data.txt')
                            save_text(fitSingleCurve.fitfunction_label, fitSingleCurveFolder + '/fitfunctionLabel.txt')
                            array2csv(fitSingleCurve.fitrange, fitSingleCurveFolder + '/fitrange.csv')
                            array2csv(fitSingleCurve.fitresult, fitSingleCurveFolder + '/fitresult.csv')
                            #save_text(fitSingleCurve.fitfunction, fitSingleCurveFolder + '/fitfunction.txt')
                            array2csv(fitSingleCurve.fitarray, fitSingleCurveFolder + '/fitarray.csv')
                            array2csv(fitSingleCurve.startvalues, fitSingleCurveFolder + '/startvalues.csv')
                            array2csv(fitSingleCurve.param_factors10, fitSingleCurveFolder + '/paramFactors10.csv')
                            array2csv(fitSingleCurve.paramidx, fitSingleCurveFolder + '/paramidx.csv')
                            save_text(check_none(fitSingleCurve.w0), fitSingleCurveFolder + '/w0.txt')
                            save_text(fitSingleCurve.D, fitSingleCurveFolder + '/D.txt')
        
        if root != 0:
            root.progressMessage = 'Almost there...'
        make_zip(fname, folderName)
        shutil.rmtree(folderName)
        

def savelib_ffs(FFSlib, root=0, fname=''):
    if fname == '':
        fname = save_ffs()

    if fname is not None:
        if root != 0:
            root.progress = 0
            root.progressMessage = 'Saving file...'
        
        with h5py.File(fname, "w") as f:
            # save version
            inf = f.create_group('info')                                                                   
            inf.attrs['version'] = check_none("1.1", dtype=str)
            inf.attrs['notes'] = check_none(FFSlib.notes, dtype=str)
            inf.attrs['active_image'] = check_none(FFSlib.active_image, dtype=int)
            inf.attrs['date_created'] = check_none(FFSlib.date_created, dtype=str)
            inf.attrs['date_modified'] = check_none(str(datetime.now()), dtype=str)
            
            lib = FFSlib.lib
            Nimages = len(lib)
            for i in range(Nimages):
                if root != 0:
                    root.progress = i / Nimages
                    root.progressMessage = 'Saving image ' + str(i+1) + '...'
                im = lib[i]
                
                # image found
                im5 = f.create_group("image" + str(i))
                im5.attrs["image_name"] = check_none(im.image_name, dtype=str)
                im5.create_dataset('image',data=im.image, compression="gzip")
                im5.create_dataset('image_raw',data=im.image, compression="gzip")
                
                # store active file
                im5.attrs["active_ffs"] = check_none(im.active_ffs, dtype=int)
                
                # go through ffs files
                Nfiles = len(im.ffs_list)
                for j in range(Nfiles):
                    if root != 0:
                        root.progress = i / Nimages + j / (Nimages * Nfiles)
                        root.progressMessage = 'Saving FFS file ' + str(j+1) + ', image ' + str(i+1) + '...'
                    file = im.ffs_list[j]
                    
                    file5 = im5.create_group("file" + str(j))
                    file5.attrs["fname"] = check_none(file.fname, dtype=str)
                    file5.attrs["label"] = check_none(file.label, dtype=str)
                    
                    # save time trace
                    tt = file.timetrace
                    if tt is None:
                        tt = np.zeros((1)) - 1
                    file5.create_dataset('timetrace', data=tt, compression="gzip")
                    
                    # save finger print
                    fp = file.airy
                    if fp is None:
                        fp = np.zeros((1)) - 1
                    file5.create_dataset('airy', data=fp, compression="gzip")
                    
                    # save active analysis
                    file5.attrs["active_analysis"] = check_none(file.active_analysis)
                    
                    # save metadata
                    md = file.metadata
                    for key in list(md.__dict__.keys()):
                        if key != 'coords':
                            file5.attrs[key] = check_none(getattr(md, key))
                        else:
                            file5.create_dataset(key, data=getattr(md, key))
                        
                    # go through each analysis
                    Nanalysis = len(file.analysis_list)
                    for k in range(Nanalysis):
                        if root != 0:
                            root.progressMessage = 'Saving analysis ' + str(k+1) +  ', FFS file ' + str(j+1) + ', image ' + str(i+1) + '...'
                        analysis = file.analysis_list[k]
                        
                        analysis5 = file5.create_group("analysis" + str(k))
                        
                        # save analysis mode
                        analysis5.attrs["mode"] = check_none(analysis.mode)
                        # save elements
                        analysis5.attrs["elements"] = check_none(analysis.settings.elements)
                        # save list of G
                        analysis5.attrs["list_of_g"] = check_none(analysis.settings.list_of_g)
                        # save average
                        analysis5.attrs["average"] = check_none(analysis.settings.average)
                        # save algorithm
                        analysis5.attrs["algorithm"] = check_none(analysis.settings.algorithm)
                        
                        # save resolution
                        analysis5.attrs["resolution"] = check_none(analysis.settings.resolution)
                        # save chunck size
                        analysis5.attrs["chunksize"] = check_none(analysis.settings.chunksize)
                        # save chunks off
                        analysis5.create_dataset("chunks_off",data=analysis.settings.chunks_off)
                        
                        # save active fit
                        analysis5.attrs["active_fit"] = check_none(analysis.active_fit)
                        # save correlations
                        c = analysis.corrs
                        if c is not None:
                            for key in list(c.__dict__.keys()):
                                if key != "dwellTime" and key != "chunks_off":
                                    # DWELLTIME IS ALREADY STORED IN METADATA
                                    analysis5.create_dataset(key, data=getattr(c, key))
        
                        # save fits
                        root.progressMessage = 'Saving fit ' + str(k+1) +  ', FFS file ' + str(j+1) + ', image ' + str(i+1) + '...'
                        fits5 = analysis5.create_group("fits")
                        fits = analysis.fits
                        Nfits = len(fits)
                        
                        root.progressMessage = 'Saving fits ' + str(k+1) +  ', FFS file ' + str(j+1) + ', image ' + str(i+1) + '...'
                        for l in range(Nfits):
                            fitAllCurves5 = fits5.create_group("fit" + str(l))
                            fitAllCurves = fits[l].fit_all_curves # contains 3 fits: central, sum3, sum5
                            Nfits3 = len(fitAllCurves)
                            for m in range(Nfits3):
                                root.progressMessage = 'Saving fit--- ' + str(m) +  ', FFS file ' + str(j+1) + ', image ' + str(i+1) + '...'
                                fitSingleCurve5 = fitAllCurves5.create_group("fit_single_curve" + str(m))
                                fitSingleCurve = fitAllCurves[m]
                                # save all fields
                                fitSingleCurve5.create_dataset('minbound',data=fitSingleCurve.minbound)
                                fitSingleCurve5.create_dataset('maxbound',data=fitSingleCurve.maxbound)
                                fitSingleCurve5.create_dataset('fitresult',data=fitSingleCurve.fitresult)
                                fitSingleCurve5.create_dataset('fitarray',data=fitSingleCurve.fitarray)
                                fitSingleCurve5.create_dataset('startvalues',data=fitSingleCurve.startvalues)
                                fitSingleCurve5.create_dataset('param_factors10',data=fitSingleCurve.param_factors10)
                                fitSingleCurve5.create_dataset('paramidx',data=fitSingleCurve.paramidx)
                                fitSingleCurve5.attrs['data'] = check_none(fitSingleCurve.data)
                                fitSingleCurve5.attrs['fitfunction_label'] = check_none(fitSingleCurve.fitfunction_label)
                                fitSingleCurve5.attrs['fitrange'] = check_none(fitSingleCurve.fitrange)
                                fitSingleCurve5.attrs['w0'] = check_none(fitSingleCurve.w0)
                                check_none(fitSingleCurve.D)
                                fD = fitSingleCurve.D
                                if fD is None:
                                    fD = 'None'
                                fitSingleCurve5.attrs['D'] = fD
                                
                                
            if root != 0:
                root.progressMessage = 'Almost there...'

def save_ffs(window_title='Save project as', ftype='*.ffs', directory=''):
    """
    Select name to save ffs lib
    ===========================================================================
    Input       Meaning
    ---------------------------------------------------------------------------
    dialog window is opened to choose file name
    ===========================================================================
    Output      Meaning
    ---------------------------------------------------------------------------
    the file name is returned
    ===========================================================================
    """
    
    dialog = QFileDialog()
    dialog.setWindowTitle(window_title)
    dialog.setNameFilter(ftype)
    dialog.setDirectory(directory)
    fname = QFileDialog.getSaveFileName()
    
    if fname[0] == '':
        return None
    return fname[0]

def make_zip(fname, zipfolder):
    if fname[-4:] != '.ffz':
        fname = fname + '.ffz'
    # make zip
    shutil.make_archive('tempfile', 'zip', zipfolder)
    # store zip as ffz
    try:
        os.rename('tempfile.zip', 'tempfile.ffz')
    except:
        os.remove('tempfile.ffz')
        os.rename('tempfile.zip', 'tempfile.ffz')
    # move ffs file do desired location
    try:
        shutil.move('tempfile.ffz', fname)
    except:
        os.remove(fname)
        shutil.move('tempfile.ffz', fname)

def unzip(fname):
    originalFilename = fname
    if originalFilename[-4:] == '.ffz':
        fname = originalFilename[:-4] + '.zip'
        os.rename(originalFilename, fname)
    randString = str(int(np.random.rand()*100000))
    folderName = "Facts_tempData_" + randString
    create_dir(folderName)
    shutil.unpack_archive(fname, folderName)
    if originalFilename[-4:] == '.ffz':
        os.rename(fname, originalFilename)
    return folderName

def create_dir(path):
    os.mkdir(path)

def check_none(inpt, dtype=None):
    if dtype is not None and inpt is not None and inpt != 'None':
        inpt = dtype(inpt)
    if inpt is None:
        inpt = 'None'
    elif type(inpt) is str and inpt == 'None':
        inpt = None
    return inpt

def to_text(text):
    if text is None:
        text = "None"
    text = str(text)
    return text

def save_image(image, fname):
    if image is not None:
        array2csv(image, fname, dtype=int)

def save_text(text, fname):
    text = to_text(text)
    with open(fname, "w") as text_file:
        text_file.write(text)

def read_text(fname, dtype=str):    
    try:
        with open(fname, "r") as text_file:
            text = text_file.read()
            if text == 'None':
                text = None
            else:
                text = dtype(text)
    except:
        text = None
    return text


# -*- coding: utf-8 -*-
"""
Created on Wed May 22 10:46:35 2019

@author: SPAD-fcs
"""

class Correlations:
    pass

def select_g(g, selection='average'):
    """
    Return a selection of the autocorrelations

    Parameters
    ----------
    g : TYPE
        Object with all autocorrelations, i.e. output of e.g.
        fcs2corrsplit.
    selection : TYPE, optional
        Default value 'average': select only the autocorrelations that 
        are averaged over multiple time traces.
        E.g. if fcs2corrsplit splits a time trace in 10 pieces,
        calculates G for each trace and then calculates the average G, 
        all autocorrelations are stored in G. This function removes all
        of them except for the average G. The default is 'average'.

    Returns
    -------
    Autocorrelation object with only the pixel dwell time and the 
    average autocorrelations stored. All other autocorrelations are
    removed.

    """
    
    # get all attributes of G
    Glist = list(g.__dict__.keys())
    
    if selection == 'average':
        # make a new list containing only 'average' attributes
        Glist2 = [s for s in Glist if "average" in s]
    else:
        Glist2 = Glist
    
    # make a new object with the average attributes
    Gout = Correlations()
    for i in Glist2:
        setattr(Gout, i, getattr(g, i))
        
    # add dwell time
    Gout.dwellTime = g.dwellTime
    
    return(Gout)



def checkfname(fname, extension):
    """
    Check if string ends with given extension

    Parameters
    ----------
    fname : string
        String with file name.
    extension : string
        Extension that fname should have.

    Returns
    -------
    fname : string
        String with the original fname and added extension if needed.

    """
    
    if fname is None:
        return None
    
    extL = len(extension)
    if len(fname) <= extL or fname[-extL:] != extension:
        fname = fname + "." + extension
    return fname



import numpy as np
import os
from numpy import genfromtxt
from .checkfname import checkfname

def csv2array(file, dlmt='\t'):
    """
    Import .csv file into 2D array

    Parameters
    ----------
    file : string
        file name.
    dlmt : string, optional
        delimiter. The default is '\t'.

    Returns
    -------
    data : np.array
        data array.

    """
    
    if dlmt == -1:
        #guess delimiter
        with open(file) as f:
            if '\t' in f.read():
                dlmt = '\t'
            else:
                dlmt = ','
    file = checkfname(file, 'csv')
    if os.path.getsize(file) == 0:
        return []
    data = genfromtxt(file, delimiter=dlmt)
    if type(data) is not np.ndarray or data.ndim < 1:
        # if it is a single number -> convert to array
        try:
            data = np.atleast_1d(data)
        except:
            pass
    return data


def array2csv(data, fname='test.csv', dlmt='\t', dtype=float):
    """
    Store 2D array into .csv file

    Parameters
    ----------
    data : np.array
        data array.
    fname : string, optional
        file name. The default is 'test.csv'.
    dlmt : string, optional
        delimiter. The default is '\t'.
    dtype : data type, optional
        data type. The default is float.

    Returns
    -------
    None.

    """
    
    data = np.asarray(data, dtype)
    fname = checkfname(fname, 'csv')
    if dtype == int:
        np.savetxt(fname, data, delimiter=dlmt, fmt='%i')
    else:
        np.savetxt(fname, data, delimiter=dlmt)


import numpy as np
import os
from .read_binfile import read_binfile
from .u64_to_counts import u64_to_counts
from ..tools.savevar import savevar


class ArrivalTimesObject:
    pass


def binfile2arrivaltimes(fname, store_pickle=False, dwellTime=[]):
    """
    Read a binary fcs data file, containing a list of U64 numbers that
    represent the photon counts per microtime bin of about 1 µs and store the
    arrival times in units of dwellTime.
    Storing arrivaltimes in general reduces the file size significantly

    Parameters
    ----------
    fname : string
        Binary file name.
    store_pickle : Boolean, optional
        Store result in Pickle file. The default is False.
    dwellTime : float, optional
        Pixel dwell time [s]. The default is [].

    Returns
    -------
    arrival_times : list
        26 lists with the arrival times of the photon for each of
        the 25 detector elements + the sum.

    """
    
    readLength = 8 * 65536

    # Get file size [in bytes]
    fsize = os.path.getsize(fname)

    # Empty object to store photon arrival times
    arrival_times = ArrivalTimesObject()
    for i in range(25):
        setattr(arrival_times, 'det' + str(i), [])
    arrival_times.sum = []
    arrival_times.dwellTime = dwellTime

    startPos = 0  # position in the .bin file to start reading
    t = 0  # iterator used to indicate current time bin

    # Go through the file, read a chunk of data, write to the object, and repeat
    while startPos < fsize:
        # Read part of the binary file
        data = read_binfile(fname, startPos, readLength)
        startPos += readLength

        # Calculate photon counts
        for i in range(len(data)):
            counts = u64_to_counts(data[i])
            # go through each detector element
            for det in range(25):
                for c in range(counts[det]):
                    getattr(arrival_times,
                            'det' + str(det)).append(int(np.floor(t/2)))
            # sum of all detector elements
            for c in range(counts[25]):
                arrival_times.sum.append(int(np.floor(t/2)))
            t += 1

    
    if store_pickle:
        print("Storing .pickle file")
        savevar(arrival_times, fname[0:-4] + "_arrival_times.pickle")
        print(".pickle file stored")

    return arrival_times
